import dataclasses
from pathlib import Path
from unittest.mock import MagicMock

import msgpack
from ymmsl import (
        Operator, Reference, Checkpoints, CheckpointRangeRule, CheckpointAtRule)

import libmuscle
from libmuscle.logging import LogLevel
from libmuscle.manager.mmp_server import MMPRequestHandler
from libmuscle.mcp.protocol import RequestType, ResponseType
from libmuscle.snapshot import SnapshotMetadata


def test_create_servicer(
        logger, profile_store, mmp_configuration, instance_registry,
        topology_store, snapshot_registry, deadlock_detector):
    MMPRequestHandler(
            logger, profile_store, mmp_configuration, instance_registry,
            topology_store, snapshot_registry, deadlock_detector, None)


def test_log_message(mmp_request_handler, caplog):
    request = [
            RequestType.SUBMIT_LOG_MESSAGE.value,
            'test_instance_id', 0.0, LogLevel.WARNING.value,
            'Testing log message']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = mmp_request_handler.handle_request(encoded_request)

    decoded_result = msgpack.unpackb(result, raw=False)

    assert isinstance(decoded_result, list)
    assert len(decoded_result) == 1
    assert decoded_result[0] == ResponseType.SUCCESS.value

    assert caplog.records[0].name == 'test_instance_id'
    assert caplog.records[0].levelname == 'WARNING'
    assert caplog.records[0].message == 'Testing log message'


def test_get_settings(mmp_configuration, mmp_request_handler):
    request = [RequestType.GET_SETTINGS.value]
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert len(decoded_result) == 2
    assert decoded_result[0] == ResponseType.SUCCESS.value
    assert decoded_result[1] == {}

    mmp_configuration.settings['test1'] = 13
    mmp_configuration.settings['test2'] = 12.3
    mmp_configuration.settings['test3'] = 'testing'
    mmp_configuration.settings['test4'] = True
    mmp_configuration.settings['test5'] = [2.3, 7.4]
    mmp_configuration.settings['test6'] = [[1.0, 2.0], [2.0, 1.0]]

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)
    assert len(decoded_result) == 2
    assert decoded_result[0] == ResponseType.SUCCESS.value

    result_dict = decoded_result[1]
    assert len(result_dict) == 6

    assert result_dict['test1'] == 13
    assert result_dict['test2'] == 12.3
    assert result_dict['test3'] == 'testing'
    assert result_dict['test4'] is True
    assert result_dict['test5'] == [2.3, 7.4]
    assert result_dict['test6'] == [[1.0, 2.0], [2.0, 1.0]]
    assert result_dict == mmp_configuration.settings.as_ordered_dict()


def test_register_instance(mmp_request_handler, instance_registry):
    request = [
            RequestType.REGISTER_INSTANCE.value,
            'test_instance',
            ['tcp://localhost:10000'],
            [['test_in', 'F_INIT']],
            libmuscle.__version__]
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.SUCCESS.value
    assert (instance_registry._locations['test_instance'] ==
            ['tcp://localhost:10000'])

    registered_ports = instance_registry._ports
    assert registered_ports['test_instance'][0].name == 'test_in'
    assert registered_ports['test_instance'][0].operator == Operator.F_INIT


def test_register_instance_no_version(mmp_request_handler):
    request = [
            RequestType.REGISTER_INSTANCE.value,
            'test_instance',
            ['tcp://localhost:10000'],
            [['test_in', 'F_INIT']]]
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.ERROR.value
    assert 'version' in decoded_result[1]


def test_register_instance_version_mismatch(mmp_request_handler):
    request = [
            RequestType.REGISTER_INSTANCE.value,
            'test_instance',
            ['tcp://localhost:10000'],
            [['test_in', 'F_INIT']],
            libmuscle.__version__ + "dev"]
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.ERROR.value
    assert 'version' in decoded_result[1]


def test_get_checkpoint_info(mmp_configuration, mmp_request_handler):
    resume_path = Path('/path/to/resume.pack')
    mmp_configuration.resume = {Reference('test_instance'): resume_path}
    mmp_configuration.checkpoints = Checkpoints(
            True,
            [CheckpointRangeRule(every=10), CheckpointAtRule([1, 2, 3.0])])

    request = [RequestType.GET_CHECKPOINT_INFO.value, 'test_instance']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.SUCCESS.value
    elapsed_time, checkpoints, resume, snapshot_directory = decoded_result[1:]

    assert elapsed_time > 0.0

    assert isinstance(checkpoints, dict)
    assert checkpoints.keys() == {'at_end', 'wallclock_time', 'simulation_time'}
    assert checkpoints['at_end'] is True
    assert checkpoints['simulation_time'] == []
    wallclock_time = checkpoints['wallclock_time']
    assert len(wallclock_time) == 2
    assert wallclock_time[0] == {'start': None, 'stop': None, 'every': 10}
    assert all(isinstance(obj, (type(None), float))
               for obj in wallclock_time[0].values())
    assert wallclock_time[1] == {'at': [1, 2, 3.0]}
    assert all(isinstance(obj, (type(None), float))
               for obj in wallclock_time[1]['at'])

    assert resume is not None
    assert Path(resume) == resume_path

    assert snapshot_directory is None


def test_get_checkpoint_info2(registered_mmp_request_handler2, tmp_path):
    request = [RequestType.GET_CHECKPOINT_INFO.value, 'test_instance']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = registered_mmp_request_handler2.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.SUCCESS.value
    snapshot_directory = decoded_result[4]
    assert snapshot_directory == (
            str(tmp_path) + '/instances/test_instance/snapshots')


def test_double_register_instance(mmp_request_handler):
    request = [
            RequestType.REGISTER_INSTANCE.value,
            'test_instance',
            ['tcp://localhost:10000'],
            [['test_in', 'F_INIT']],
            libmuscle.__version__]
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.SUCCESS.value

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.ERROR.value
    assert 'test_instance' in decoded_result[1]


def test_deregister_instance(
        registered_mmp_request_handler, instance_registry):
    assert Reference('macro') in instance_registry._locations

    request = [RequestType.DEREGISTER_INSTANCE.value, 'macro']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = registered_mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.SUCCESS.value
    assert Reference('macro') not in instance_registry._locations


def test_get_peers_pending(mmp_request_handler):
    request = [RequestType.GET_PEERS.value, 'micro[0][0]']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.PENDING.value


def test_request_peers_fanout(registered_mmp_request_handler):
    request = [RequestType.GET_PEERS.value, 'macro']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = registered_mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    status, conduits, dims, locations = decoded_result
    assert status == ResponseType.SUCCESS.value

    assert conduits[0][0] == 'macro.out'
    assert conduits[0][1] == 'micro.in'
    assert conduits[1][0] == 'micro.out'
    assert conduits[1][1] == 'macro.in'

    assert dims['micro'] == [10, 10]

    for i, (name, locs) in enumerate(locations.items()):
        assert name == f'micro[{i // 10}][{i % 10}]'
        assert locs == [f'direct:{name}']


def test_request_peers_fanin(registered_mmp_request_handler):
    request = [RequestType.GET_PEERS.value, 'micro[4][3]']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = registered_mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    status, conduits, dims, locations = decoded_result
    assert status == ResponseType.SUCCESS.value

    assert conduits[0][0] == 'macro.out'
    assert conduits[0][1] == 'micro.in'
    assert conduits[1][0] == 'micro.out'
    assert conduits[1][1] == 'macro.in'

    assert dims['macro'] == []

    assert locations['macro'] == ['direct:macro']


def test_request_peers_bidir(registered_mmp_request_handler2):
    request = [RequestType.GET_PEERS.value, 'meso[2]']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = registered_mmp_request_handler2.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    status, conduits, dims, locations = decoded_result
    assert status == ResponseType.SUCCESS.value

    assert conduits[0][0] == 'macro.out'
    assert conduits[0][1] == 'meso.in'
    assert conduits[1][0] == 'meso.out'
    assert conduits[1][1] == 'micro.in'
    assert conduits[2][0] == 'micro.out'
    assert conduits[2][1] == 'meso.in'
    assert conduits[3][0] == 'meso.out'
    assert conduits[3][1] == 'macro.in'

    assert dims['micro'] == [5, 10]
    assert dims['macro'] == []

    assert locations['macro'] == ['direct:macro']

    assert locations['macro'] == ['direct:macro']
    for i in range(10):
        assert locations[f'micro[2][{i}]'] == [f'direct:micro[2][{i}]']


def test_request_peers_own_conduits(registered_mmp_request_handler2):
    request = [RequestType.GET_PEERS.value, 'macro']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = registered_mmp_request_handler2.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    status, conduits, dims, locations = decoded_result
    assert status == ResponseType.SUCCESS.value

    assert conduits[0][0] == 'macro.out'
    assert conduits[0][1] == 'meso.in'
    assert conduits[1][0] == 'meso.out'
    assert conduits[1][1] == 'macro.in'


def test_request_peers_unknown(registered_mmp_request_handler2):
    request = [RequestType.GET_PEERS.value, 'does_not_exist']
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = registered_mmp_request_handler2.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    status, error_msg = decoded_result
    assert status == ResponseType.ERROR.value
    assert error_msg is not None
    assert 'does_not_exist' in error_msg


def test_submit_snapshot(registered_mmp_request_handler):
    register_snapshot = MagicMock()
    registered_mmp_request_handler._snapshot_registry.register_snapshot = \
        register_snapshot

    instance_id = 'micro[1][2]'
    snapshot = SnapshotMetadata(
            ['1', '2'], 1.234, 2.345, 3.456,
            {'in': [1], 'out': [0]}, True, 'fname')
    snapshot_dict = dataclasses.asdict(snapshot)

    request = [RequestType.SUBMIT_SNAPSHOT.value, instance_id, snapshot_dict]
    encoded_request = msgpack.packb(request, use_bin_type=True)

    result = registered_mmp_request_handler.handle_request(encoded_request)
    decoded_result = msgpack.unpackb(result, raw=False)

    assert decoded_result[0] == ResponseType.SUCCESS.value
    register_snapshot.assert_called_once_with(Reference(instance_id), snapshot)



from typing import Optional

from ymmsl import Reference

from libmuscle.mcp.transport_client import TimeoutHandler
from libmuscle.mmp_client import MMPClient


class Deadlock(Exception):
    """Exception that is raised when the simulation has deadlocked."""


class ReceiveTimeoutHandler(TimeoutHandler):
    """Timeout handler when receiving messages from peers.

    This handler sends a message to the Muscle Manager when the receive times out (and
    another message when the message does arrive).

    This is used by the manager to detect if the simulation is in a deadlock, where a
    cycle of instances is waiting for each other.
    """

    def __init__(
            self, manager: MMPClient,
            peer_instance: Reference, port_name: str, slot: Optional[int],
            timeout: float
            ) -> None:
        """Initialize a new timeout handler.

        Args:
            manager: Connection to the muscle manager.
            peer_instance: the peer instance we try to receive from.
            port_name: the name of the port we try to receive on.
            slot: the slot we try to receive on.
            timeout: Timeout in seconds.
        """
        self._manager = manager
        self._peer_instance = peer_instance
        self._port_name = port_name
        self._slot = slot
        self._timeout = timeout
        # Counter to keep track of the number of timeouts
        self._num_timeouts = 0

    @property
    def timeout(self) -> float:
        # Increase timeout by a factor 1.5 with every timeout we hit:
        factor = 1.5 ** self._num_timeouts
        return self._timeout * factor

    def on_timeout(self) -> None:
        if self._num_timeouts == 0:
            # Notify the manager that we're waiting for a receive
            self._manager.waiting_for_receive(
                    self._peer_instance, self._port_name, self._slot)
        else:
            # Ask the manager if we're part of a detected deadlock
            if self._manager.is_deadlocked():
                raise Deadlock()
        self._num_timeouts += 1

    def on_receive(self) -> None:
        self._manager.waiting_for_receive_done(
                self._peer_instance, self._port_name, self._slot)



from random import uniform
from threading import Condition, Lock, Thread
import time
from typing import List

from libmuscle.mmp_client import MMPClient
from libmuscle.profiling import ProfileEvent, ProfileTimestamp


_COMMUNICATION_INTERVAL = 10.0  # seconds


class Profiler:
    """Collects profiling events and sends them to the manager.
    """
    def __init__(self, manager: MMPClient) -> None:
        """Create a Profiler.

        Args:
            manager: The client used to submit data to the manager.
        """
        # Protects all member variables and _flush()
        self._mutex = Lock()

        self._manager = manager
        self._enabled = True
        self._events: List[ProfileEvent] = []
        self._thread = Thread(target=self._communicate, daemon=True)
        self._done_cv = Condition(self._mutex)
        self._done = False
        self._next_send = 0.0

        self._thread.start()

    def shutdown(self) -> None:
        with self._mutex:
            if self._done:
                return

            self._done = True
            self._done_cv.notify_all()

        self._thread.join()

        # with the thread gone, there's no need to lock anymore
        self._flush()

    def set_level(self, level: str) -> None:
        """Set the detail level at which data is collected.

        Args:
            level: Either 'none' or 'all' to disable or enable sending
                    events to the manager.
        """
        with self._mutex:
            self._enabled = level == 'all'

    def record_event(self, event: ProfileEvent) -> None:
        """Record a profiling event.

        This will record the event, and may flush this and previously
        recorded events to the manager. If the time is still running,
        it will be stopped. Other than this the event must be complete
        when it is submitted. Do not use the event object after calling
        this function with it.

        Args:
            event: The event to record.
        """
        if event.stop_time is None:
            event.stop_time = ProfileTimestamp()

        if self._enabled:
            with self._mutex:
                self._events.append(event)
                if len(self._events) >= 10000:
                    self._flush()
                self._next_send = time.monotonic() + _COMMUNICATION_INTERVAL

    def _communicate(self) -> None:
        """Background thread that communicates with the manager.

        This runs in the background, and periodically sends events to
        the manager.
        """
        initial_delay = uniform(0.0, _COMMUNICATION_INTERVAL)

        with self._mutex:
            self._next_send = time.monotonic() + initial_delay

            while not self._done:
                now = time.monotonic()
                notified = self._done_cv.wait(self._next_send - now)
                if not notified:
                    now = time.monotonic()
                    if self._next_send <= now:
                        self._flush()
                        self._next_send = now + _COMMUNICATION_INTERVAL

    def _flush(self) -> None:
        """Send events to the manager and empty the queue.

        Make sure to lock self._mutex before calling this.
        """
        if self._events:
            self._manager.submit_profile_events(self._events)
            self._events.clear()



from enum import Enum


class RequestType(Enum):
    """Identifier for different types of request

    The MUSCLE Communication Protocol (MCP) defines a simple Remote Procedure
    Call protocol in which a request is sent to the server and a response is
    sent back to the calling client. In MCP, both of these are chunks of bytes.

    The MUSCLE Manager Protocol, MUSCLE Peer Protocol and MUSCLE Agent Protocol
    define the encoded messages sent in those chunks, using MsgPack encoding.
    To distinguish different kinds of requests, a request type identifier is
    used, as represented by this class.
    """
    # MUSCLE Manager Protocol
    REGISTER_INSTANCE = 1
    GET_PEERS = 2
    DEREGISTER_INSTANCE = 3
    GET_SETTINGS = 4
    SUBMIT_LOG_MESSAGE = 5
    SUBMIT_PROFILE_EVENTS = 6
    SUBMIT_SNAPSHOT = 7
    GET_CHECKPOINT_INFO = 8
    # Connection deadlock detection
    WAITING_FOR_RECEIVE = 9
    WAITING_FOR_RECEIVE_DONE = 10
    IS_DEADLOCKED = 11

    # MUSCLE Peer Protocol
    GET_NEXT_MESSAGE = 21

    # MUSCLE Agent Protocol
    REPORT_RESOURCES = 41
    GET_COMMAND = 42
    REPORT_RESULT = 43


class ResponseType(Enum):
    """Identifier for different types of response

    MCP Responses may be of different kinds, as described by this class.
    These are currently only used by the MUSCLE Manager Protocol.
    """
    # MUSCLE Manager Protocol
    SUCCESS = 0
    ERROR = 1
    PENDING = 2


class AgentCommandType(Enum):
    """Identifier for different types of commands

    These are requested from the manager by the agent, and tell it what to do. Part
    of the MUSCLE Agent Protocol, used in the response to RequestType.GET_COMMAND.
    """
    START = 1
    CANCEL_ALL = 2
    SHUTDOWN = 3



import datetime
import time
from typing import Optional


class Timestamp:
    """A timestamp, as the number of seconds since the UNIX epoch.

    Args:
        seconds: The number of seconds since the start of 1970.
    """
    def __init__(self, seconds: Optional[float] = None) -> None:
        """Create a Timestamp representing the given time, or now.

        If seconds is None, the current time is used.
        """
        if seconds is None:
            seconds = time.time()
        self.seconds = seconds

    def to_asctime(self) -> str:
        """Converts a Timestamp to a LogRecord asctime.

        Returns:
            The timestamp as a string in the format used by default
            by Python's logging subsystem.
        """
        date_time = datetime.datetime.fromtimestamp(self.seconds)
        whole_part = date_time.strftime('%Y-%m-%d %H:%M:%S')
        return '%s,%03d' % (whole_part, date_time.time().microsecond / 1000)



from typing import List
from libmuscle.peer_info import PeerInfo
from libmuscle.port_manager import PortManager

from ymmsl import Conduit, Identifier, Operator, Reference as Ref

import pytest


@pytest.fixture
def index() -> List[int]:
    return [13]


@pytest.fixture
def port_manager(index) -> PortManager:
    declared_ports = {
            Operator.O_I: ['out'],
            Operator.S: ['in']}
    return PortManager(index, declared_ports)


@pytest.fixture
def index2() -> List[int]:
    return []


@pytest.fixture
def port_manager2(index2) -> PortManager:
    declared_ports = {
            Operator.F_INIT: ['in[]'],
            Operator.O_F: ['out[]']}
    port_manager = PortManager(index2, declared_ports)

    component_id = Ref('other')
    conduits = [Conduit('component.out', 'other.in'),
                Conduit('other.out', 'component.in')]
    peer_dims = {Ref('component'): [20]}
    peer_locations = {Ref('component'): ['direct:test']}
    peer_info = PeerInfo(component_id, index2, conduits, peer_dims, peer_locations)

    port_manager.connect_ports(peer_info)
    return port_manager


def test_connect_ports(index, port_manager) -> None:
    component_id = Ref('component')
    conduits = [Conduit('component.out', 'other.in'),
                Conduit('other.settings_out', 'component.muscle_settings_in'),
                Conduit('other.out', 'component.in')]
    peer_dims = {Ref('other'): []}
    peer_locations = {Ref('other'): ['direct:test']}
    peer_info = PeerInfo(component_id, index, conduits, peer_dims, peer_locations)

    port_manager.connect_ports(peer_info)

    # check automatic ports
    assert port_manager.settings_in_connected()
    assert port_manager._muscle_settings_in.name == Identifier('muscle_settings_in')
    assert port_manager._muscle_settings_in.operator == Operator.F_INIT
    assert port_manager._muscle_settings_in._length is None

    # check declared ports
    ports = port_manager._ports
    assert ports['in'].name == Identifier('in')
    assert ports['in'].operator == Operator.S
    assert ports['in']._length is None

    assert ports['out'].name == Identifier('out')
    assert ports['out'].operator == Operator.O_I
    assert ports['out']._length is None


def test_connect_vector_ports(index) -> None:
    declared_ports = {
            Operator.F_INIT: ['in[]'],
            Operator.O_F: ['out1', 'out2[]']}

    port_manager = PortManager(index, declared_ports)

    component_id = Ref('component')
    conduits = [Conduit('other1.out', 'component.in'),
                Conduit('component.out1', 'other.in'),
                Conduit('component.out2', 'other3.in')]
    peer_dims = {
            Ref('other1'): [20, 7],
            Ref('other'): [25],
            Ref('other3'): [20]}
    peer_locations = {
            Ref('other'): ['direct:test'],
            Ref('other1'): ['direct:test1'],
            Ref('other3'): ['direct:test3']}
    peer_info = PeerInfo(component_id, index, conduits, peer_dims, peer_locations)

    port_manager.connect_ports(peer_info)

    ports = port_manager._ports
    assert ports['in'].name == Identifier('in')
    assert ports['in'].operator == Operator.F_INIT
    assert ports['in']._length == 7
    assert ports['in']._is_resizable is False

    assert ports['out1'].name == Identifier('out1')
    assert ports['out1'].operator == Operator.O_F
    assert ports['out1']._length is None

    assert ports['out2'].name == Identifier('out2')
    assert ports['out2'].operator == Operator.O_F
    assert ports['out2']._length == 0
    assert ports['out2']._is_resizable is True


def test_connect_multidimensional_ports() -> None:
    index = [13]
    declared_ports = {Operator.F_INIT: ['in[][]']}
    port_manager = PortManager(index, declared_ports)

    component_id = Ref('component')
    conduits = [Conduit(Ref('other.out'), Ref('component.in'))]
    peer_dims = {Ref('other'): [20, 7, 30]}
    peer_locations = {Ref('other'): ['direct:test']}
    peer_info = PeerInfo(component_id, index, conduits, peer_dims, peer_locations)

    with pytest.raises(ValueError):
        port_manager.connect_ports(peer_info)


def test_connect_inferred_ports() -> None:
    index = [13]
    port_manager = PortManager(index, None)

    component_id = Ref('component')
    conduits = [Conduit('other1.out', 'component.in'),
                Conduit('component.out1', 'other.in'),
                Conduit('component.out3', 'other2.in')]
    peer_dims = {
            Ref('other1'): [20, 7],
            Ref('other'): [25],
            Ref('other2'): []}
    peer_locations = {
            Ref('other'): ['direct:test'],
            Ref('other1'): ['direct:test1'],
            Ref('other2'): ['direct:test2']}
    peer_info = PeerInfo(component_id, index, conduits, peer_dims, peer_locations)

    port_manager.connect_ports(peer_info)

    ports = port_manager._ports
    assert ports['in'].name == Identifier('in')
    assert ports['in'].operator == Operator.F_INIT
    assert ports['in']._length == 7
    assert ports['in']._is_resizable is False

    assert ports['out1'].name == Identifier('out1')
    assert ports['out1'].operator == Operator.O_F
    assert ports['out1']._length is None

    assert ports['out3'].name == Identifier('out3')
    assert ports['out3'].operator == Operator.O_F
    assert ports['out3']._length is None


def test_port_message_counts(port_manager) -> None:
    port_manager.connect_ports(PeerInfo('component', [], [], {}, {}))

    msg_counts = port_manager.get_message_counts()
    assert msg_counts == {'in': [0], 'out': [0], 'muscle_settings_in': [0]}

    port_manager.get_port('in').increment_num_messages()

    msg_counts2 = port_manager.get_message_counts()
    assert msg_counts2 == {'in': [1], 'out': [0], 'muscle_settings_in': [0]}

    port_manager.get_port('out').increment_num_messages()
    port_manager.get_port('out').increment_num_messages()

    msg_counts3 = port_manager.get_message_counts()
    assert msg_counts3 == {'in': [1], 'out': [2], 'muscle_settings_in': [0]}

    port_manager._muscle_settings_in.increment_num_messages()

    msg_counts4 = port_manager.get_message_counts()
    assert msg_counts4 == {'in': [1], 'out': [2], 'muscle_settings_in': [1]}

    port_manager.restore_message_counts(msg_counts)

    msg_counts5 = port_manager.get_message_counts()
    assert msg_counts5 == {'in': [0], 'out': [0], 'muscle_settings_in': [0]}

    with pytest.raises(RuntimeError):
        port_manager.restore_message_counts({"x?invalid_port": 3})


def test_vector_port_message_counts(port_manager2) -> None:
    msg_counts = port_manager2.get_message_counts()
    assert msg_counts == {'out': [0] * 20,
                          'in': [0] * 20,
                          'muscle_settings_in': [0]}

    port_manager2.get_port('out').increment_num_messages(13)

    msg_counts = port_manager2.get_message_counts()
    assert msg_counts == {'out': [0] * 13 + [1] + [0] * 6,
                          'in': [0] * 20,
                          'muscle_settings_in': [0]}

    port_manager2.restore_message_counts({'out': list(range(20)),
                                          'in': list(range(20)),
                                          'muscle_settings_in': [4]})

    port_manager2.get_port('out').increment_num_messages(13)

    msg_counts = port_manager2.get_message_counts()
    assert msg_counts == {'out': list(range(13)) + [14] + list(range(14, 20)),
                          'in': list(range(20)),
                          'muscle_settings_in': [4]}



import os

from libmuscle import Instance


def component() -> None:
    """A component that checks whether the environment is correct.

    This prints the values of the BASHRC_LOADED and MANAGER_SHELL
    variables so that the test can see whether they're set correctly.

    It also checks if slurm is on the path, which it only is if the
    default modules are still loaded, at least in this particular
    test setup, and whether we have VIRTUAL_ENV set.

    """
    instance = Instance()

    print(os.environ.get('PROFILE_LOADED', '0'))
    print(os.environ.get('BASHRC_LOADED', '0'))
    print(os.environ.get('MANAGER_SHELL', '0'))
    print('1' if 'slurm' in os.environ['PATH'] else '0')
    print(os.environ.get('VIRTUAL_ENV', ''))

    while instance.reuse_instance():
        pass


if __name__ == '__main__':
    component()



import pytest

from libmuscle.manager.deadlock_detector import DeadlockDetector


@pytest.fixture
def detector() -> DeadlockDetector:
    return DeadlockDetector()


def test_no_deadlock(detector: DeadlockDetector) -> None:
    detector.waiting_for_receive("macro", "micro", "s", None)
    assert not detector.is_deadlocked("macro")
    assert not detector.is_deadlocked("micro")
    detector.waiting_for_receive_done("macro", "micro", "s", None)
    assert not detector.is_deadlocked("macro")
    assert not detector.is_deadlocked("micro")


def test_double_waiting_log_error(detector: DeadlockDetector) -> None:
    detector.waiting_for_receive("macro", "micro", "s", 0)
    with pytest.raises(AssertionError):
        detector.waiting_for_receive("macro", "micro", "s", 1)


def test_not_waiting_log_error(detector: DeadlockDetector) -> None:
    with pytest.raises(AssertionError):
        detector.waiting_for_receive_done("macro", "micro", "s", 0)


def test_waiting_for_different_instance_log_error(detector: DeadlockDetector) -> None:
    detector.waiting_for_receive("macro", "micro", "s", 0)
    with pytest.raises(AssertionError):
        detector.waiting_for_receive_done("macro", "meso", "s", 0)


def test_waiting_for_different_port_log_error(detector: DeadlockDetector) -> None:
    detector.waiting_for_receive("macro", "micro", "s", 0)
    with pytest.raises(AssertionError):
        detector.waiting_for_receive_done("macro", "micro", "f_init", 0)


def test_deadlock(detector: DeadlockDetector) -> None:
    detector.waiting_for_receive("macro", "micro", "s", None)
    assert not detector.is_deadlocked("macro")
    assert not detector.is_deadlocked("micro")
    detector.waiting_for_receive("micro", "macro", "f_init", None)
    assert detector.is_deadlocked("macro")
    assert detector.is_deadlocked("micro")


def test_deadlock_cancelled(detector: DeadlockDetector) -> None:
    detector.waiting_for_receive("macro", "micro", "s", None)
    detector.waiting_for_receive("micro", "macro", "f_init", None)
    detector.waiting_for_receive_done("macro", "micro", "s", None)
    assert not detector.is_deadlocked("macro")
    assert not detector.is_deadlocked("micro")


def test_double_deadlock(detector: DeadlockDetector) -> None:
    detector.waiting_for_receive("macro", "micro", "s", None)
    detector.waiting_for_receive("micro", "macro", "f_init", None)
    detector.waiting_for_receive("cycle2", "peer2", "s", None)
    detector.waiting_for_receive("peer2", "cycle2", "f_init", None)
    detector.waiting_for_receive_done("macro", "micro", "s", None)
    assert not detector.is_deadlocked("macro")
    assert not detector.is_deadlocked("micro")
    assert detector.is_deadlocked("cycle2")
    assert detector.is_deadlocked("peer2")



import logging
import sys
import types
from typing import List, Optional

from libmuscle.port_manager import PortManager
from ymmsl import Operator


_logger = logging.getLogger(__name__)


class MMSFValidator:
    """The MMSF Validator checks whether Instances are following the Multiscale
    Modelling and Simulation Framework when sending and receiving messages.

    In particular it checks that in order:

    1. reuse_instance() is called
    2. Messages are received on all F_INIT ports
    3. The following sub-items happen in order, 0 or more times:

        a. Messages are sent on all O_I ports
        b. Messages are received on all S ports

    4. Messages are sent on all O_F ports

    If any message is sent or received out of order, a warning is logged to indicate
    that the instance is not following the MMSF pattern. In some cases (for example the
    time bridge in ``examples/python/interact_coupling.py``) this is expected and the
    warnings can be disabled by setting the
    :attr:`~libmuscle.instance.InstanceFlags.SKIP_MMSF_SEQUENCE_CHECKS` flag.

    Note:
        Checks on vector ports are not implemented. When the instance uses vector ports,
        the MMSF Validator will be disabled.
    """
    def __init__(self, port_manager: PortManager) -> None:
        self._port_manager = port_manager

        port_names = port_manager.list_ports()
        port_objects = {
                operator: [port_manager.get_port(name) for name in names]
                for operator, names in port_names.items()}
        self._connected_ports = {
                operator: [str(port.name) for port in ports if port.is_connected()]
                for operator, ports in port_objects.items()}
        self._port_operators = {
                port: operator
                for operator, ports in port_names.items()
                for port in ports}

        if self._connected_ports.get(Operator.NONE, []):
            _logger.warning(
                    "This instance is using ports with Operator.NONE. This does not "
                    "adhere to the Multiscale Modelling and Simulation Framework "
                    "and may lead to deadlocks. You can disable this warning by "
                    "setting the flag InstanceFlags.SKIP_MMSF_SEQUENCE_CHECKS "
                    "when creating the libmuscle.Instance.")

        # Allowed operator transitions, the following are unconditionally allowed:
        self._allowed_transitions = {
                Operator.NONE: [Operator.NONE, Operator.F_INIT],
                Operator.F_INIT: [Operator.O_I, Operator.O_F],
                Operator.O_I: [Operator.S],
                Operator.S: [Operator.O_I, Operator.O_F],
                Operator.O_F: [Operator.NONE]}
        # If there are operators without connected ports, we can skip over those
        # This logic is transitive, i.e. when there are no connected ports for both
        # F_INIT and O_I, we will also add NONE -> S to self._allowed_transition:
        for operator in [Operator.F_INIT, Operator.O_I, Operator.S, Operator.O_F]:
            if not self._connected_ports.get(operator, []):
                # Find all transitions A -> operator -> B and allow transition A -> B:
                for from_op, allowed in self._allowed_transitions.items():
                    if from_op is operator:
                        continue
                    if operator not in allowed:
                        continue
                    for to_op in self._allowed_transitions[operator]:
                        if to_op not in allowed:
                            allowed.append(to_op)
        # Sort allowed transitions for more logical log messages
        for allowed in self._allowed_transitions.values():
            allowed.sort(key=lambda op: op.value)

        # Disable this validator when the instance uses vector ports to keep this class
        # simpler. Support for vector ports may be added in the future.
        self._enabled = not any(
                port.is_vector() for ports in port_objects.values() for port in ports)
        if self._enabled:
            _logger.debug("MMSF Validator is enabled")
        else:
            _logger.debug(
                    "MMSF Validator is disabled: this instance uses vector ports, "
                    "which are not supported by the MMSF Validator.")

        # State tracking
        self._current_ports_used: List[str] = []
        self._current_operator: Operator = Operator.NONE

    def check_send(self, port_name: str, slot: Optional[int]) -> None:
        """Check that sending on the provided port adheres to the MMSF."""
        self._check_send_receive(port_name, slot)

    def check_receive(self, port_name: str, slot: Optional[int]) -> None:
        """Check that receiving on the provided port adheres to the MMSF."""
        self._check_send_receive(port_name, slot)

    def reuse_instance(self) -> None:
        """Check that a reuse_instance() adheres to the MMSF."""
        if not self._enabled:
            return
        self._check_transition(Operator.NONE)

    def skip_f_init(self) -> None:
        """Call when resuming from an intermediate snapshot: F_INIT is skipped."""
        # Pretend we're now in F_INIT and we have already received on all F_INIT ports:
        self._current_operator = Operator.F_INIT
        self._current_ports_used = self._connected_ports.get(Operator.F_INIT, [])

    def _check_send_receive(
            self, port_name: str, slot: Optional[int]) -> None:
        """Actual implementation of check_send/check_receive."""
        if not self._enabled:
            return

        operator = self._port_operators[port_name]
        if self._current_operator != operator:
            # Operator changed, check that all ports were used in the previous operator
            self._check_transition(operator, port_name)

        if port_name in self._current_ports_used:
            # We're using the same port for a second time, this is fine if:
            # 1. We're allowed to do this operator immediately again, and
            # 2. All ports of the current operator have been used
            # Both are checked by _check_transition:
            self._check_transition(operator, port_name)

        self._current_ports_used.append(port_name)

    def _check_transition(self, operator: Operator, port_name: str = "") -> None:
        """Check that a transition to the provided operator is allowed.

        Log a warning when the transition does not adhere to the MMSF.

        Args:
            operator: Operator to transition to.
            port_name: The name of the port that was sent/receveived on. This is only
                used for constructing the warning message.
        """
        connected_ports = self._connected_ports.get(self._current_operator, [])
        expected: str = ""

        unused_ports = [
                port for port in connected_ports
                if port not in self._current_ports_used]
        if unused_ports:
            # We didn't complete the current phase
            if self._current_operator.allows_receiving():
                expected = "a receive"
            else:
                expected = "a send"
            expected += (
                    f" on any of these {self._current_operator.name} ports: "
                    + ", ".join(unused_ports))

        elif operator not in self._allowed_transitions[self._current_operator]:
            # Transition to the operator is not allowed.
            # Build the message we want to display to users:
            expected_lst = []
            for to_op in self._allowed_transitions[self._current_operator]:
                ports = ', '.join(map(repr, self._connected_ports.get(to_op, [])))
                if to_op is Operator.NONE:
                    expected_lst.append("a call to reuse_instance()")
                elif not ports:
                    continue
                elif to_op.allows_receiving():
                    expected_lst.append(f"a receive on an {to_op.name} port ({ports})")
                else:
                    expected_lst.append(f"a send on an {to_op.name} port ({ports})")
            assert expected_lst
            expected = " or ".join(expected_lst)

        if expected:
            # We expected something else, log a warning:
            if operator is Operator.NONE:
                action = "reuse_instance()"
            elif operator in (Operator.F_INIT, Operator.S):
                action = f"Receive on port '{port_name}'"
            else:
                action = f"Send on port '{port_name}'"

            # Find the file:line where the user called send/receive/reuse_instance
            try:
                frame: Optional[types.FrameType] = sys._getframe()
            except Exception:
                frame = None  # sys._getframe() is not guaranteed available
            while (frame
                    and frame.f_globals.get("__name__", "").startswith("libmuscle.")):
                # This frame is still part of a libmuscle module, step up:
                frame = frame.f_back
            loc = f" ({frame.f_code.co_filename}:{frame.f_lineno})" if frame else ""

            _logger.warning(
                "%s%s does not adhere to the MMSF: was expecting %s.\n"
                "Not adhering to the Multiscale Modelling and Simulation Framework "
                "may lead to deadlocks. You can disable this warning by "
                "setting the flag InstanceFlags.SKIP_MMSF_SEQUENCE_CHECKS "
                "when creating the libmuscle.Instance.",
                action, loc, expected)

        self._current_operator = operator
        self._current_ports_used = []



from enum import IntEnum
from typing import Any, cast, Optional

import msgpack
import numpy as np

from ymmsl import Reference, Settings

from libmuscle.grid import Grid


class ExtTypeId(IntEnum):
    """MessagePack extension type ids.

    MessagePack lets you define your own types as an extension to the
    built-in ones. These are distinguished by a number from 0 to 127.
    This class is our registry of extension type ids.
    """
    CLOSE_PORT = 0
    SETTINGS = 1
    GRID_INT32 = 2
    GRID_INT64 = 3
    GRID_FLOAT32 = 4
    GRID_FLOAT64 = 5
    GRID_BOOL = 6


_grid_types = {
        ExtTypeId.GRID_INT32,
        ExtTypeId.GRID_INT64,
        ExtTypeId.GRID_FLOAT32,
        ExtTypeId.GRID_FLOAT64,
        ExtTypeId.GRID_BOOL}


class ClosePort:
    """Sentinel value to send when closing a port.

    Sending an object of this class on a port/conduit conveys to the
    receiver the message that no further messages will be sent on this
    port during the simulation.

    All information is carried by the type, this has no attributes.
    """
    pass


def _encode_grid(grid: Grid) -> msgpack.ExtType:
    """Encodes a Grid object into the wire format.
    """
    ext_type_map = {
            'int32': ExtTypeId.GRID_INT32,
            'int64': ExtTypeId.GRID_INT64,
            'float32': ExtTypeId.GRID_FLOAT32,
            'float64': ExtTypeId.GRID_FLOAT64,
            'bool': ExtTypeId.GRID_BOOL}

    array = grid.array
    if array.flags.f_contiguous:
        # indexes that differ in the first place are adjacent
        order = 'fa'
    else:
        # indexes that differ in the last place are adjacent
        order = 'la'

    # dtype is a bit weird, but this seems to be consistent
    if isinstance(array.dtype, np.dtype):
        array_type = str(array.dtype)
    else:
        array_type = str(np.dtype(array_type))

    if array_type not in ext_type_map:
        raise RuntimeError('Unsupported array data type')

    buf = array.tobytes(order='A')

    # array_type is redundant, but useful metadata.
    grid_dict = {
            'type': array_type,
            'shape': list(array.shape),
            'order': order,
            'data': buf,
            'indexes': grid.indexes}
    packed_data = msgpack.packb(grid_dict, use_bin_type=True)
    return msgpack.ExtType(ext_type_map[array_type], packed_data)


def _decode_grid(code: int, data: bytes) -> Grid:
    """Creates a Grid from serialised data.
    """
    type_map = {
            ExtTypeId.GRID_INT32: np.int32,
            ExtTypeId.GRID_INT64: np.int64,
            ExtTypeId.GRID_FLOAT32: np.float32,
            ExtTypeId.GRID_FLOAT64: np.float64,
            ExtTypeId.GRID_BOOL: np.bool_}

    order_map = {
            'fa': 'F',
            'la': 'C'}

    grid_dict = msgpack.unpackb(data, raw=False)
    order = order_map[grid_dict['order']]
    shape = tuple(grid_dict['shape'])
    dtype = type_map[ExtTypeId(code)]
    array = np.ndarray(shape, dtype, grid_dict['data'], order=order)  # type: ignore
    indexes = grid_dict['indexes']
    if indexes == []:
        indexes = None
    return Grid(array, indexes)


def _data_encoder(obj: Any) -> Any:
    """Encodes custom objects for MessagePack.

    In particular, this takes care of any Settings, Grid and
    numpy.ndarray objects the user may want to send.
    """
    if isinstance(obj, ClosePort):
        return msgpack.ExtType(ExtTypeId.CLOSE_PORT, bytes())
    elif isinstance(obj, Settings):
        packed_data = msgpack.packb(obj.as_ordered_dict(),
                                    use_bin_type=True)
        return msgpack.ExtType(ExtTypeId.SETTINGS, packed_data)
    elif isinstance(obj, np.ndarray):
        return _encode_grid(Grid(obj))
    elif isinstance(obj, Grid):
        return _encode_grid(obj)
    return obj


def _ext_decoder(code: int, data: bytes) -> msgpack.ExtType:
    if code == ExtTypeId.CLOSE_PORT:
        return ClosePort()
    elif code == ExtTypeId.SETTINGS:
        plain_dict = msgpack.unpackb(data, raw=False)
        return Settings(plain_dict)
    elif code in _grid_types:
        return _decode_grid(code, data)
    return msgpack.ExtType(code, data)


class MPPMessage:
    """A MUSCLE Peer Protocol message.

    Messages carry the identity of their sender and receiver, so that
    they can be routed by a MUSCLE Transport Overlay when we get to
    multi-site running in the future.
    """
    def __init__(self, sender: Reference, receiver: Reference,
                 port_length: Optional[int],
                 timestamp: float, next_timestamp: Optional[float],
                 settings_overlay: Settings, message_number: int,
                 saved_until: float, data: Any
                 ) -> None:
        """Create an MPPMessage.

        Senders and receivers are refered to by a Reference, which
        contains Instance[InstanceNumber].Port[Slot].

        The port_length field is only used if two vector ports are
        connected together. In that case the number of slots is not
        determined by the number of instances, and must be set by
        the sender and then communicated to the receiver in this
        additional field in all messages sent on the port.

        Args:
            sender: The sending endpoint.
            receiver: The receiving endpoint.
            port_length: Length of the slot, where applicable.
            settings_overlay: The serialised overlay settings.
            message_number: Sequence number on this conduit.
            saved_until: Elapsed time until which the sender has
                processed checkpoints.
            data: The serialised contents of the message.
        """
        # make sure timestamp and next_timestamp are floats
        timestamp = float(timestamp)
        if next_timestamp is not None:
            next_timestamp = float(next_timestamp)

        self.sender = sender
        self.receiver = receiver
        self.port_length = port_length
        self.timestamp = timestamp
        self.next_timestamp = next_timestamp
        self.settings_overlay = settings_overlay
        self.message_number = message_number
        self.saved_until = saved_until
        if isinstance(data, np.ndarray):
            self.data = Grid(data)
        else:
            self.data = data

    @staticmethod
    def from_bytes(message: bytes) -> 'MPPMessage':
        """Create an MPP Message from an encoded buffer.

        Args:
            message: MessagePack encoded message data.
        """
        message_dict = msgpack.unpackb(
                message, ext_hook=_ext_decoder, raw=False)
        sender = Reference(message_dict["sender"])
        receiver = Reference(message_dict["receiver"])
        port_length = message_dict["port_length"]
        timestamp = message_dict["timestamp"]
        next_timestamp = message_dict["next_timestamp"]
        settings_overlay = message_dict["settings_overlay"]
        message_number = message_dict["message_number"]
        saved_until = message_dict["saved_until"]

        data = message_dict["data"]
        return MPPMessage(
                sender, receiver, port_length, timestamp, next_timestamp,
                settings_overlay, message_number, saved_until, data)

    def encoded(self) -> bytes:
        """Encode the message and return as a bytes buffer.
        """
        message_dict = {
                'sender': str(self.sender),
                'receiver': str(self.receiver),
                'port_length': self.port_length,
                'timestamp': self.timestamp,
                'next_timestamp': self.next_timestamp,
                'settings_overlay': self.settings_overlay,
                'message_number': self.message_number,
                'saved_until': self.saved_until,
                'data': self.data
                }

        return cast(bytes, msgpack.packb(
            message_dict, default=_data_encoder, use_bin_type=True))



class RequestHandler:
    """Handles requests sent to a TransportServer.

    TransportServers operate in terms of chunks of bytes received and
    sent in return. RequestHandlers interpret received chunks of bytes,
    handle the request, and return a chunk of bytes containing an
    encoded response.
    """
    def handle_request(self, request: bytes) -> bytes:
        """Handle a request.

        Args:
            request: A received request

        Returns:
            An encoded response
        """
        raise NotImplementedError()     # pragma: no cover

    def close(self) -> None:
        """Free per-thread resources.

        On shutdown of the server, this will be called by each server
        thread before it shuts down.
        """
        pass


class ServerNotSupported(RuntimeError):
    pass


class TransportServer:
    """A server that accepts MCP connections.

    This is a base class for MCP Servers. An MCP Server accepts
    connections over some lower-level communication protocol, receives
    requests and returns responses from a RequestHandler.
    """
    def __init__(self, handler: RequestHandler) -> None:
        """Create a TransportServer.

        Args:
            handler: A handler to handle requests.
        """
        self._handler = handler

    def get_location(self) -> str:
        """Returns the location this server listens on.

        Returns:
            A string containing the location.
        """
        raise NotImplementedError()  # pragma: no cover

    def close(self) -> None:
        """Closes this server.

        Stops the server listening, waits for existing clients to
        disconnect, then frees any other resources.
        """
        raise NotImplementedError()  # pragma: no cover



import logging
from unittest.mock import MagicMock, Mock, patch

import pytest

from libmuscle.communicator import Communicator, Message
from libmuscle.mpp_message import ClosePort, MPPMessage
from libmuscle.peer_info import PeerInfo
from ymmsl import Conduit, Reference as Ref, Settings


@pytest.fixture
def profiler():
    return MagicMock()


@pytest.fixture(autouse=True)
def MPPServer():
    with patch('libmuscle.communicator.MPPServer') as MPPServer:
        yield MPPServer


@pytest.fixture
def mpp_server(MPPServer):
    return MPPServer.return_value


@pytest.fixture
def port_manager():
    with patch('libmuscle.communicator.PortManager') as PortManager:
        port_manager = PortManager.return_value
        port_manager.settings_in_connected.return_value = False
        yield port_manager


@pytest.fixture(autouse=True)
def MPPClient():
    with patch('libmuscle.communicator.MPPClient') as MPPClient:
        yield MPPClient


@pytest.fixture
def mpp_client(MPPClient):
    return MPPClient.return_value


@pytest.fixture
def communicator(connected_port_manager, profiler):
    return Communicator(Ref('component'), [], connected_port_manager, profiler, Mock())


@pytest.fixture
def connected_communicator(communicator):
    # These work with declared_ports and connected_port_manager in conftest.py
    conduits = [
            Conduit('peer.out', 'component.in'),
            Conduit('peer2.out_v', 'component.in_v'),
            Conduit('peer3.out_r', 'component.in_r'),
            Conduit('component.out_v', 'peer2.in'),
            Conduit('component.out_r', 'peer3.in_r'),
            Conduit('component.out', 'peer.in')]

    peer_dims = {Ref('peer'): [], Ref('peer2'): [13], Ref('peer3'): []}

    peer_locations = {
            Ref('peer'): ['tcp:peer:9001'], Ref('peer3'): ['tcp:peer3:9001']}
    peer_locations.update({
        Ref(f'peer2[{s}]'): ['tcp:peer2:9001'] for s in range(13)})

    peer_info = PeerInfo(Ref('component'), [], conduits, peer_dims, peer_locations)
    communicator.set_peer_info(peer_info)
    return communicator


def test_create_communicator(communicator, mpp_server):
    assert communicator._server == mpp_server
    pass


def test_send_message(connected_communicator, mpp_server):
    msg = Message(0.0, 1.0, 'Testing', Settings({'s0': 0, 's1': '1'}))

    connected_communicator.send_message('out_v', msg, 7, -1.0)

    mpp_server.deposit.assert_called_once()
    args = mpp_server.deposit.call_args[0]
    assert args[0] == Ref('peer2[7].in')

    encoded_msg = MPPMessage.from_bytes(args[1])
    assert encoded_msg.sender == Ref('component.out_v[7]')
    assert encoded_msg.receiver == Ref('peer2[7].in')
    assert encoded_msg.port_length is None
    assert encoded_msg.timestamp == 0.0
    assert encoded_msg.next_timestamp == 1.0
    assert len(encoded_msg.settings_overlay) == 2
    assert encoded_msg.settings_overlay['s0'] == 0
    assert encoded_msg.settings_overlay['s1'] == '1'
    assert encoded_msg.message_number == 0
    assert encoded_msg.saved_until == -1.0
    assert encoded_msg.data == 'Testing'


def test_send_message_disconnected(connected_communicator, mpp_server):
    msg = MagicMock()

    connected_communicator.send_message('not_connected', msg)

    mpp_server.deposit.assert_not_called()


def test_receive_message(connected_communicator, mpp_client):
    msg = MPPMessage(
            Ref('peer.out'), Ref('component.in'), None, 2.0, 3.0,
            Settings({'s0': '0', 's1': True}), 0, 1.0, 'Testing')

    mpp_client.receive.return_value = msg.encoded(), MagicMock()

    connected_communicator.set_receive_timeout(-1)
    recv_msg, saved_until = connected_communicator.receive_message('in')

    mpp_client.receive.assert_called_with(Ref('component.in'), None)

    assert recv_msg.timestamp == 2.0
    assert recv_msg.next_timestamp == 3.0
    assert recv_msg.data == 'Testing'
    assert len(recv_msg.settings) == 2
    assert recv_msg.settings['s0'] == '0'
    assert recv_msg.settings['s1'] is True
    assert saved_until == 1.0


def test_receive_message_vector(connected_communicator, mpp_client):
    msg = MPPMessage(
            Ref('peer2.out_v'), Ref('component.in_v'), 5, 4.0, 6.0,
            Settings({'s0': [0.0], 's1': 1.0}), 0, 3.5, 'Testing2')

    mpp_client.receive.return_value = msg.encoded(), MagicMock()

    connected_communicator.set_receive_timeout(-1)
    recv_msg, saved_until = connected_communicator.receive_message('in_v', 5)

    mpp_client.receive.assert_called_with(Ref('component.in_v[5]'), None)

    assert recv_msg.timestamp == 4.0
    assert recv_msg.next_timestamp == 6.0
    assert recv_msg.data == 'Testing2'
    assert len(recv_msg.settings) == 2
    assert recv_msg.settings['s0'] == [0.0]
    assert recv_msg.settings['s1'] == 1.0
    assert saved_until == 3.5


def test_receive_close_port(connected_communicator, mpp_client, port_manager):
    msg = MPPMessage(
            Ref('peer.out'), Ref('component.in'), None, float('inf'), None,
            Settings(), 0, 0.1, ClosePort())

    mpp_client.receive.return_value = msg.encoded(), MagicMock()

    recv_msg, saved_until = connected_communicator.receive_message('in')

    assert port_manager.get_port('in').is_open() is False


def test_receive_close_port_vector(connected_communicator, mpp_client, port_manager):
    msg = MPPMessage(
            Ref('peer2.out_v'), Ref('component.in_v'), 5, float('inf'), None,
            Settings(), 0, 3.5, ClosePort())

    mpp_client.receive.return_value = msg.encoded(), MagicMock()

    recv_msg, saved_until = connected_communicator.receive_message('in_v', 5)

    assert port_manager.get_port('in_v').is_open(5) is False


def test_port_count_validation(
        connected_communicator, mpp_client, connected_port_manager):

    msg = MPPMessage(
            Ref('peer.out'), Ref('component.in'),
            None, 0.0, None, Settings({'test1': 12}), 0, 7.6,
            b'test')

    mpp_client.receive.return_value = msg.encoded(), MagicMock()

    connected_communicator.receive_message('in')
    assert connected_port_manager.get_port('in').get_message_counts() == [1]

    with pytest.raises(RuntimeError):
        # the message received has message_number = 0 again
        connected_communicator.receive_message('in')


def test_port_discard_error_on_resume(
        caplog, connected_communicator, mpp_client, connected_port_manager):

    msg = MPPMessage(
            Ref('other.out[13]'), Ref('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}), 1, 2.3,
            b'test')

    mpp_client.receive.return_value = msg.encoded(), MagicMock()

    connected_port_manager.get_port('out').restore_message_counts([0])
    connected_port_manager.get_port('in').restore_message_counts([2])

    for port in ('out', 'in'):
        assert connected_port_manager.get_port(port)._is_resuming == [True]
        assert connected_port_manager.get_port(port).is_resuming(None)

    # In the next block, the first message with message_number=1 is discarded.
    # The RuntimeError is raised when 'receiving' the second message with
    # message_number=1
    with caplog.at_level(logging.DEBUG, 'libmuscle.communicator'):
        with pytest.raises(RuntimeError):
            connected_communicator.receive_message('in')

        assert any([
            'Discarding received message' in rec.message
            for rec in caplog.records])


def test_port_discard_success_on_resume(
        caplog, connected_communicator, mpp_client, connected_port_manager):

    side_effect = [(MPPMessage(
            Ref('other.out[13]'), Ref('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}), message_number, 1.0,
            {'this is message': message_number}).encoded(), MagicMock())
            for message_number in [1, 2]]

    mpp_client.receive.side_effect = side_effect

    connected_port_manager.get_port('out').restore_message_counts([0])
    connected_port_manager.get_port('in').restore_message_counts([2])

    for port in ('out', 'in'):
        assert connected_port_manager.get_port(port)._is_resuming == [True]
        assert connected_port_manager.get_port(port).is_resuming(None)

    with caplog.at_level(logging.DEBUG, 'libmuscle.communicator'):
        msg, _ = connected_communicator.receive_message('in')
        assert any([
            'Discarding received message' in rec.message
            for rec in caplog.records])

    # message_number=1 should have been discarded:
    assert msg.data == {'this is message': 2}
    assert connected_communicator._port_manager.get_port(
            'in').get_message_counts() == [3]


def test_shutdown(
        connected_communicator, mpp_client, connected_port_manager, mpp_server):

    msg = MPPMessage(
            Ref('peer.out'), Ref('component.in'), None, float('inf'), None,
            Settings(), 0, 0.0, ClosePort())

    messages = {Ref('component.in'): msg}

    port_sender = {
            'in_v': 'peer2[x].out_v',
            'in_r': 'peer3.out_r[x]'}

    for port_name, snd in port_sender.items():
        port = connected_port_manager.get_port(port_name)
        for slot in range(port.get_length()):
            sender = Ref(snd.replace('x', str(slot)))
            receiver = Ref(f'component.{port_name}[{slot}]')

            messages[receiver] = MPPMessage(
                    sender, receiver, slot, float('inf'), None,
                    Settings(), 0, 3.5, ClosePort())

    def receive(receiver, timeout_handler):
        return messages[receiver].encoded(), MagicMock()

    mpp_client.receive = receive

    connected_communicator.shutdown()

    expected_receivers = (
            {Ref('peer.in')} |
            {
                Ref(f'peer2[{slot}].in')
                for slot in range(
                    connected_port_manager.get_port('out_v').get_length())} |
            {
                Ref(f'peer3.in[{slot}]')
                for slot in range(
                    connected_port_manager.get_port('out_r').get_length())})

    for call in mpp_server.deposit.call_args_list:
        assert call[0][0] in expected_receivers
        msg = MPPMessage.from_bytes(call[0][1])
        assert isinstance(msg.data, ClosePort)
        expected_receivers.remove(call[0][0])

    assert not expected_receivers



from typing import List

from ymmsl import Identifier, Reference


class Endpoint:
    """Place that a message is sent from and to.

    In the model description, we have kernels with ports connected by
    conduits. However, these kernels may be replicated, in which case
    there are many instances of them at run time. Thus, at run time
    there also need to be many conduit instances to connect the many
    kernel instances.

    A conduit always connects a port on a kernel to another port on
    another kernel. A conduit instance connects an endpoint to another
    endpoint. An endpoint has the name of a kernel, its index, the name
    of a port on that kernel, and a *slot*. The kernel and port name
    of a sender or receiver of a conduit instance come from the
    corresponding conduit.

    When a kernel is instantiated multiple times, the instances each
    have a unique index, which is a list of integers, to distinguish
    them from each other. Since a conduit instance connects kernel
    instances, each side will have an index to supply to the endpoint.
    The slot is an optional integer, like the index, and is passed when
    sending or receiving a message, and gives additional information
    on where to send the message.

    For example, assume a single kernel named ``abc`` with port ``p1``
    which is connected to a port ``p2`` on kernel ``def`` by a conduit,
    and of kernel ``def`` there are 10 instances. A message sent by
    ``abc`` on ``p1`` to the fourth instance of ``def`` port ``p2`` is
    sent from an endpoint with kernel ``abc``, index ``[]``, port
    ``p1`` and slot ``3``, and received on an endpoint with kernel
    ``def``, index ``[3]``, port ``p2`` and slot ``None``.

    Conduit instances are never actually created in the code, but
    Endpoints are.
    """
    def __init__(self, kernel: Reference, index: List[int], port: Identifier,
                 slot: List[int]) -> None:
        """Create an Endpoint

        Note: kernel is a Reference, not an Identifier, because it may
        have namespace parts.

        Args:
            kernel: Name of an instance's kernel.
            index: Index of the kernel instance.
            port: Name of the port used.
            slot: Slot on which to send or receive.
        """
        self.kernel = kernel
        self.index = index
        self.port = port
        self.slot = slot

    def ref(self) -> Reference:
        """Express as Reference.

        This yields a valid Reference of the form
        kernel[index].port[slot], with index and port omitted if they
        are zero-length.

        Returns:
            A Reference to this Endpoint.
        """
        ret = self.kernel
        if self.index:
            ret += self.index
        ret += self.port
        if self.slot:
            ret += self.slot
        return ret

    def __str__(self) -> str:
        """Convert to string.

        Returns this Endpoint as the string for of a Reference to it.
        See :meth:ref().

        Returns:
            The string representation of this Endpoint.
        """
        return str(self.ref())

    def instance(self) -> Reference:
        """Get a Reference to the instance this endpoint is on.
        """
        ret = self.kernel
        if self.index:
            ret += self.index
        return ret



import errno
import logging
import time
from typing import Any, Dict, cast, List, Optional

import msgpack
from ymmsl import (
        Conduit, Identifier, Operator, Port, Reference, PartialConfiguration,
        Checkpoints, CheckpointRule, CheckpointAtRule, CheckpointRangeRule)

import libmuscle
from libmuscle.logging import LogLevel
from libmuscle.manager.instance_registry import (
        AlreadyRegistered, InstanceRegistry)
from libmuscle.manager.logger import Logger
from libmuscle.manager.run_dir import RunDir
from libmuscle.manager.snapshot_registry import SnapshotRegistry
from libmuscle.manager.topology_store import TopologyStore
from libmuscle.mcp.protocol import RequestType, ResponseType
from libmuscle.mcp.tcp_transport_server import TcpTransportServer
from libmuscle.mcp.transport_server import RequestHandler
from libmuscle.manager.profile_store import ProfileStore
from libmuscle.manager.deadlock_detector import DeadlockDetector
from libmuscle.profiling import (
        ProfileEvent, ProfileEventType, ProfileTimestamp)
from libmuscle.snapshot import SnapshotMetadata
from libmuscle.timestamp import Timestamp


_logger = logging.getLogger(__name__)


def decode_operator(data: str) -> Operator:
    """Create an Operator from a MsgPack-compatible value."""
    return Operator[data]


def decode_port(data: List[str]) -> Port:
    """Create a Port from a MsgPack-compatible value."""
    return Port(Identifier(data[0]), decode_operator(data[1]))


def encode_conduit(conduit: Conduit) -> List[str]:
    """Convert a Conduit to a MsgPack-compatible value."""
    return [str(conduit.sender), str(conduit.receiver)]


def encode_checkpoint_rule(rule: CheckpointRule) -> Dict[str, Any]:
    """Convert a CheckpointRule to a MsgPack-compatible value."""
    if isinstance(rule, CheckpointAtRule):
        return {'at': list(map(float, rule.at))}
    if isinstance(rule, CheckpointRangeRule):
        return {
            'start': None if rule.start is None else float(rule.start),
            'stop': None if rule.stop is None else float(rule.stop),
            'every': float(rule.every)}
    raise TypeError(f"Unknown checkpoint rule type: {type(rule)}.")


def encode_checkpoints(checkpoints: Checkpoints) -> Dict[str, Any]:
    """Convert a Checkpoins to a MsgPack-compatible value."""
    return {
        "at_end": checkpoints.at_end,
        "wallclock_time":
            list(map(encode_checkpoint_rule, checkpoints.wallclock_time)),
        "simulation_time":
            list(map(encode_checkpoint_rule, checkpoints.simulation_time)),
    }


class MMPRequestHandler(RequestHandler):
    """Handles Manager requests."""
    def __init__(
            self,
            logger: Logger,
            profile_store: ProfileStore,
            configuration: PartialConfiguration,
            instance_registry: InstanceRegistry,
            topology_store: TopologyStore,
            snapshot_registry: SnapshotRegistry,
            deadlock_detector: DeadlockDetector,
            run_dir: Optional[RunDir]
            ) -> None:
        """Create an MMPRequestHandler.

        Args:
            logger: The Logger component to log messages to.
            settings: The global settings to serve to instances.
            instance_registry: The database for instances.
            topology_store: Keeps track of how to connect things.
        """
        self._logger = logger
        self._profile_store = profile_store
        self._configuration = configuration
        self._instance_registry = instance_registry
        self._topology_store = topology_store
        self._snapshot_registry = snapshot_registry
        self._deadlock_detector = deadlock_detector
        self._run_dir = run_dir
        self._reference_time = time.monotonic()

    def handle_request(self, request: bytes) -> bytes:
        """Handles a manager request.

        Args:
            request: The encoded request

        Returns:
            response: An encoded response
        """
        req_list = msgpack.unpackb(request, raw=False)
        req_type = req_list[0]
        req_args = req_list[1:]
        if req_type == RequestType.REGISTER_INSTANCE.value:
            response = self._register_instance(*req_args)
        elif req_type == RequestType.GET_PEERS.value:
            response = self._get_peers(*req_args)
        elif req_type == RequestType.DEREGISTER_INSTANCE.value:
            response = self._deregister_instance(*req_args)
        elif req_type == RequestType.GET_SETTINGS.value:
            response = self._get_settings(*req_args)
        elif req_type == RequestType.SUBMIT_LOG_MESSAGE.value:
            response = self._submit_log_message(*req_args)
        elif req_type == RequestType.SUBMIT_PROFILE_EVENTS.value:
            response = self._submit_profile_events(*req_args)
        elif req_type == RequestType.SUBMIT_SNAPSHOT.value:
            response = self._submit_snapshot(*req_args)
        elif req_type == RequestType.GET_CHECKPOINT_INFO.value:
            response = self._get_checkpoint_info(*req_args)
        elif req_type == RequestType.WAITING_FOR_RECEIVE.value:
            response = self._waiting_for_receive(*req_args)
        elif req_type == RequestType.WAITING_FOR_RECEIVE_DONE.value:
            response = self._waiting_for_receive_done(*req_args)
        elif req_type == RequestType.IS_DEADLOCKED.value:
            response = self._is_deadlocked(*req_args)

        return cast(bytes, msgpack.packb(response, use_bin_type=True))

    def close(self) -> None:
        """Free per-thread resources.

        On shutdown of the server, this will be called by each server
        thread before it shuts down.
        """
        self._profile_store.close()

    def _register_instance(
            self, instance_id: str, locations: List[str],
            ports: List[List[str]], version: str = '') -> Any:
        """Handle a register instance request.

        Args:
            instance_id: ID of the instance to register
            locations: Locations where it can be reached
            ports: Ports of this instance
            version: Version of libmuscle that this instance uses

        Returns:
            A list containing the following values:

            status (ResponseType): SUCCESS or ERROR
            error_msg (str): An error message, only present if status
                equals ERROR
        """
        if version != libmuscle.__version__:
            return [
                    ResponseType.ERROR.value,
                    f'Instance libmuscle version ({version}) does not match'
                    f' manager libmuscle version ({libmuscle.__version__}).'
                    ' Please ensure that the instance and the manager use the'
                    ' same version of libmuscle.']

        port_objs = [decode_port(p) for p in ports]
        instance = Reference(instance_id)
        try:
            self._instance_registry.add(instance, locations, port_objs)

            _logger.info(f'Registered instance {instance_id}')
            return [ResponseType.SUCCESS.value]
        except AlreadyRegistered:
            return [
                    ResponseType.ERROR.value,
                    f'An instance with name {instance_id} was already'
                    ' registered. Did you start a non-MPI component using'
                    ' mpirun?']

    def _get_peers(self, instance_id: str) -> Any:
        """Handle a get peers request.

        Args:
            instance_id: ID of the instance requesting peers

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
            conduits (List[List[str]]): Conduits from/to peers
            dimensions (Dict[str, List[int]]): Dimensions of peer
                components
            locations (Dict[str, List[str]]): Locations where peer
                instances can be contacted.

            Or the following values on error:

            status (ResponseType): ERROR
            error_msg (str): An error message

            Or the following values if the result is not yet available:

            status (ResponseType): PENDING
            status_msg (str): A message on what we're waiting for.
        """
        # get info from yMMSL
        instance = Reference(instance_id)
        component = instance.without_trailing_ints()
        if not self._topology_store.has_kernel(component):
            return [ResponseType.ERROR.value, f'Unknown component {component}']

        conduits = self._topology_store.get_conduits(component)
        mmp_conduits = [encode_conduit(c) for c in conduits]

        peer_dims = self._topology_store.get_peer_dimensions(component)
        mmp_dimensions = {str(name): dims for name, dims in peer_dims.items()}

        # generate instances
        try:
            peers = self._topology_store.get_peer_instances(instance)
            instance_locations = {
                    str(peer): self._instance_registry.get_locations(peer)
                    for peer in peers}
        except KeyError as e:
            return [
                    ResponseType.PENDING.value,
                    f'Waiting for component {e.args[0]}']

        _logger.debug(f'Sent peers to {instance_id}')
        return [
                ResponseType.SUCCESS.value,
                mmp_conduits, mmp_dimensions, instance_locations]

    def _deregister_instance(self, instance_id: str) -> Any:
        """Handle a deregister instance request.

        Args:
            instance_id: ID of the instance to deregister

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS

            Or the following if an error occurred:

            status (ResponseType): ERROR
            error_msg (str): An error message
        """
        try:
            self._instance_registry.remove(Reference(instance_id))
            _logger.info(f'Deregistered instance {instance_id}')
            return [ResponseType.SUCCESS.value]
        except ValueError:
            return [
                    ResponseType.ERROR.value,
                    f'No instance with name {instance_id} was registered']

    def _get_settings(self) -> Any:
        """Handle a get settings request.

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
            settings (Dict[str, SettingValue]): The global settings
        """
        return [
                ResponseType.SUCCESS.value,
                self._configuration.settings.as_ordered_dict()]

    def _submit_log_message(
            self, instance_id: str, timestamp: float, level: int, text: str
            ) -> Any:
        """Handle a submit log message request.

        Args:
            instance_id: Sending instance
            timestamp: Time since epoch of the logged event
            level: Log level of the message
            text: Message text

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
        """
        self._logger.log_message(
                instance_id, Timestamp(timestamp), LogLevel(level), text)
        return [ResponseType.SUCCESS.value]

    def _submit_profile_events(
            self, instance_id: str, events: List[List[Any]]) -> Any:
        """Handle a submit profile events request.

        Args:
            instance_id: Instance that sent these events
            events: Profiling events to store

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
        """
        ev = [
                ProfileEvent(
                    ProfileEventType(e[0]), ProfileTimestamp(e[1]),
                    ProfileTimestamp(e[2]),
                    Port(e[3][0], Operator[e[3][1]]) if e[3] else None,
                    e[4], e[5], e[6], e[7], e[8])
                for e in events]

        self._profile_store.add_events(Reference(instance_id), ev)
        return [ResponseType.SUCCESS.value]

    def _submit_snapshot(
            self, instance_id: str, snapshot: Dict[str, Any]) -> Any:
        """Handle a submit snapshot request.

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
        """
        snapshot_obj = SnapshotMetadata(**snapshot)
        instance = Reference(instance_id)
        self._snapshot_registry.register_snapshot(instance, snapshot_obj)
        return [ResponseType.SUCCESS.value]

    def _get_checkpoint_info(self, instance_id: str) -> Any:
        """Get checkpoint info for an instance

        Args:
            instance: The instance whose checkpoint info to get

        Returns:
            A list containing the following values on success:

            status (ResponseType): SUCCESS
            wallclock_reference_time (float): Unix timestamp (in UTC) indicating
                wallclock time of the start of the workflow.
            checkpoints (dict): Dictionary encoding a ymmsl.Checkpoints object.
            resume_path (Optional[str]): Checkpoint filename to resume from.
            snapshot_directory (Optional[str]): Directory to store instance
                snapshots.
        """
        instance = Reference(instance_id)
        resume = None
        if instance in self._configuration.resume:
            resume = str(self._configuration.resume[instance])

        snapshot_directory = None
        if self._run_dir is not None:
            snapshot_directory = str(self._run_dir.snapshot_dir(instance))

        return [ResponseType.SUCCESS.value,
                time.monotonic() - self._reference_time,
                encode_checkpoints(self._configuration.checkpoints),
                resume,
                snapshot_directory]

    def _waiting_for_receive(
            self, instance_id: str, peer_instance_id: str,
            port_name: str, slot: Optional[int]) -> Any:
        """Indicate that the instance is waiting to receive a message.

        Args:
            instance_id: The instance that is waiting
            port_name: Port name that the instance is waiting for
            slot: Slot that the instance is waiting for
        """
        self._deadlock_detector.waiting_for_receive(
                instance_id, peer_instance_id, port_name, slot)
        return [ResponseType.SUCCESS.value]

    def _waiting_for_receive_done(
            self, instance_id: str, peer_instance_id: str,
            port_name: str, slot: Optional[int]) -> Any:
        """Indicate that the instance is done waiting to receive a message.

        Args:
            instance_id: The instance that is waiting
            port_name: Port name that the instance is waiting for
            slot: Slot that the instance is waiting for
        """
        self._deadlock_detector.waiting_for_receive_done(
                instance_id, peer_instance_id, port_name, slot)
        return [ResponseType.SUCCESS.value]

    def _is_deadlocked(self, instance_id: str) -> Any:
        """Check if the provided instance is part of a detected deadlock.
        """
        result = self._deadlock_detector.is_deadlocked(instance_id)
        return [ResponseType.SUCCESS.value, result]


class MMPServer:
    """The MUSCLE Manager Protocol server.

    This class accepts connections from the instances comprising
    the multiscale model to be executed, and services them using an
    MMPRequestHandler.
    """
    def __init__(
            self,
            logger: Logger,
            profile_store: ProfileStore,
            configuration: PartialConfiguration,
            instance_registry: InstanceRegistry,
            topology_store: TopologyStore,
            snapshot_registry: SnapshotRegistry,
            deadlock_detector: DeadlockDetector,
            run_dir: Optional[RunDir]
            ) -> None:
        """Create an MMPServer.

        This starts a TCP Transport server and connects it to an
        MMPRequestHandler, which uses the given components to service
        the requests. By default, we listen on port 9000, unless it's
        not available in which case we use a random other one.

        Args:
            logger: Logger to send log messages to
            profile_store: ProfileStore to store profile data in
            configuration: Configuration component to get settings, checkpoints
                and resumes from
            instance_registry: To register instances with and get
                peer locations from
            topology_store: To get peers and conduits from
            snapshot_registry: To register snapshots with
            run_dir: To save snapshots to
        """
        self._handler = MMPRequestHandler(
                logger, profile_store, configuration, instance_registry,
                topology_store, snapshot_registry, deadlock_detector, run_dir)
        try:
            self._server = TcpTransportServer(self._handler, 9000)
        except OSError as e:
            if e.errno != errno.EADDRINUSE:
                raise
            self._server = TcpTransportServer(self._handler)

    def get_location(self) -> str:
        """Returns this server's network location.

        This is a string of the form tcp:<hostname>:<port>.
        """
        return self._server.get_location()

    def stop(self) -> None:
        """Stops the server.

        This makes the server stop serving requests, and shuts down its
        background threads.
        """
        self._server.close()



from typing import List, Tuple

from libmuscle.planner.resources import OnNodeResources


class IAgentManager:
    """Interface for Agent Managers.

    Only implemented by AgentManager, and only exists to avoid a circular dependency
    between AgentManager, MAPServer, and MAPRequestHandler. Ugh.
    """
    def report_resources(self, resources: OnNodeResources) -> None:
        """Report resources found on a node.

        Called by MAPServer from a server thread.

        Args:
            node_name: Id of the node these resources are on
            resources: Dict mapping resource type to resource ids
        """
        raise NotImplementedError()

    def report_result(self, names_exit_codes: List[Tuple[str, int]]) -> None:
        """Report results of finished processes.

        Called by MAPServer from a server thread.

        Args:
            names_exit_codes: A list of names and exit codes of finished processes.
        """
        raise NotImplementedError()



from threading import Lock
import time
from typing import Dict

from ymmsl import Reference

from libmuscle.outbox import Outbox


class PostOffice:
    """A PostOffice is an object that holds messages to be retrieved.

    A PostOffice holds outboxes with messages for receivers. It also
    acts as a request handler for incoming requests for messages.
    """
    def __init__(self) -> None:
        """Create a PostOffice.
        """
        self._outboxes: Dict[Reference, Outbox] = {}

        self._outbox_lock = Lock()

    def have_message(self, receiver: Reference) -> bool:
        """Return whether there's a message for the given receiver.

        Args:
            receiver: The receiver of the message.
        """
        self._ensure_outbox_exists(receiver)
        return not self._outboxes[receiver].is_empty()

    def get_message(self, receiver: Reference) -> bytes:
        """Get a message from a receiver's outbox.

        Used by servers to get messages that have been sent to another
        instance.

        Args:
            receiver: The receiver of the message.
        """
        self._ensure_outbox_exists(receiver)
        return self._outboxes[receiver].retrieve()

    def deposit(self, receiver: Reference, message: bytes) -> None:
        """Deposits a message into an outbox.

        Args:
            receiver: Receiver of the message.
            message: The message to deposit.
        """
        self._ensure_outbox_exists(receiver)
        self._outboxes[receiver].deposit(message)

    def wait_for_receivers(self) -> None:
        """Waits until all outboxes are empty.
        """
        for outbox in self._outboxes.values():
            while not outbox.is_empty():
                time.sleep(0.1)

    def _ensure_outbox_exists(self, receiver: Reference) -> None:
        """Ensure that an outbox exists.

        Outboxes are created dynamically, the first time a message is
        sent to a receiver. This function checks that an outbox exists
        for a receiver, and if not, creates one.

        Args:
            receiver: The receiver that should have an outbox.
        """
        self._outbox_lock.acquire()
        if receiver not in self._outboxes:
            self._outboxes[receiver] = Outbox()
        self._outbox_lock.release()



import logging
import os
from pathlib import Path
from tempfile import TemporaryDirectory
import time

import cerulean
import pytest


_logger = logging.getLogger(__name__)


IMAGE_NAME = 'muscle3_test_cluster'

REMOTE_SHARED = '/home/cerulean/shared'

IDX_SLURM_VERSIONS = list(enumerate([
    '17-02', '17-11', '18-08', '19-05', '20-02', '20-11', '21-08', '22-05', '23-02',
    '23-11', '24-05', '24-11'
    ]))

# Shut down the containers after running the tests. Set to False to debug.
CLEAN_UP_CONTAINERS = True


skip_unless_cluster = pytest.mark.skipif(
        'MUSCLE_TEST_CLUSTER' not in os.environ,
        reason='Cluster tests were not explicitly enabled')


def run_cmd(term, timeout, command):
    exit_code, out, err = term.run(timeout, command, [])
    if exit_code != 0:
        _logger.error(err)
    assert exit_code == 0
    return out


@pytest.fixture(scope='session')
def local_term():
    return cerulean.LocalTerminal()


@pytest.fixture(scope='session')
def local_fs():
    return cerulean.LocalFileSystem()


@pytest.fixture(scope='session')
def repo_root(local_fs):
    root_dir = Path(__file__).parents[2]
    return local_fs / str(root_dir)


@pytest.fixture(scope='session')
def fake_cluster_image(local_term):
    run_cmd(local_term, 5400, (
        f'docker buildx build -t {IMAGE_NAME}'
        ' -f integration_test/fake_cluster/Dockerfile .'))


@pytest.fixture(scope='session')
def fake_cluster_image_old(local_term):
    run_cmd(local_term, 5400, (
        f'docker buildx build -t {IMAGE_NAME}_old'
        ' -f integration_test/fake_cluster/old.Dockerfile .'))


def _image_name(slurm_version):
    if slurm_version <= '20-02':
        return IMAGE_NAME + '_old'
    return IMAGE_NAME


def _gcc_version(slurm_version):
    if slurm_version <= '20-02':
        return '7.5.0'
    return '11.4.0'


def ssh_term(port, timeout_msg):
    cred = cerulean.PasswordCredential('cerulean', 'kingfisher')
    ready = False
    start = time.monotonic()
    while not ready:
        if (time.monotonic() - start) > 60.0:
            raise Exception(timeout_msg)

        try:
            term = cerulean.SshTerminal('localhost', port, cred)
            ready = True
        except Exception:
            time.sleep(3.0)

    return term


@pytest.fixture(scope='session')
def shared_dir():
    # Note that pytest's tmp_path is function-scoped, so cannot be used here
    with TemporaryDirectory(ignore_cleanup_errors=True) as tmp_dir:
        path = Path(tmp_dir)
        path.chmod(0o1777)
        yield path


@pytest.fixture(scope='session')
def cleanup_docker(local_term):
    for _, slurm_version in IDX_SLURM_VERSIONS:
        _clean_up_base_cluster(local_term, slurm_version)


def _create_network(local_term, slurm_version):
    name = f'muscle3-net-{slurm_version}'
    run_cmd(local_term, 60, f'docker network create {name}')
    return name


def _start_nodes(local_term, slurm_version, net_name, shared_dir):
    for i in range(5):
        node_name = f'node-{i}'
        hostname = f'{node_name}.example.org'

        image_name = _image_name(slurm_version)

        run_cmd(local_term, 60, (
            f'docker run -d --name={node_name}-{slurm_version} --hostname={hostname}'
            f' --network={net_name} --cap-add=CAP_SYS_NICE'
            f' --env SLURM_VERSION={slurm_version}'
            f' --mount type=bind,source={shared_dir},target={REMOTE_SHARED}'
            f' {image_name}'))


def _start_headnode(local_term, slurm_version, net_name, shared_dir, headnode_port):
    image_name = _image_name(slurm_version)

    run_cmd(local_term, 60, (
        f'docker run -d --name=headnode-{slurm_version} --hostname=headnode'
        f' --network={net_name} -p {headnode_port}:22'
        f' --env SLURM_VERSION={slurm_version}'
        f' --mount type=bind,source={shared_dir},target={REMOTE_SHARED}'
        f' {image_name}'))

    ssh_term(headnode_port, 'Virtual cluster container start timed out').close()


def _start_base_cluster(local_term, idx_slurm_version, shared_dir):
    slurm_index, slurm_version = idx_slurm_version

    headnode_port = 10022 + slurm_index

    net_name = _create_network(local_term, slurm_version)
    _start_nodes(local_term, slurm_version, net_name, shared_dir)
    _start_headnode(local_term, slurm_version, net_name, shared_dir, headnode_port)

    term = ssh_term(headnode_port, 'Connection to virtual cluster container timed out')
    fs = cerulean.SftpFileSystem(term, False)

    return term, fs, headnode_port


def _install_remote_source(local_term, repo_root, remote_fs, slurm_version):
    muscle3_tgt = remote_fs / 'home' / 'cerulean' / 'muscle3'
    muscle3_tgt.mkdir()

    container = f'headnode-{slurm_version}'

    for f in (
            'muscle3', 'libmuscle', 'scripts', 'docs', 'setup.py', 'Makefile',
            'MANIFEST.in', 'LICENSE', 'NOTICE', 'VERSION', 'README.rst'):
        run_cmd(local_term, 60, (
            f'docker cp {repo_root / f} {container}:{muscle3_tgt / f}'))

    # needs to run as root, so not run through remote_term
    run_cmd(local_term, 60, (
        f'docker exec {container} /bin/bash -c'
        f' "chown -R cerulean:cerulean {muscle3_tgt}"'))

    return muscle3_tgt


def _create_muscle3_venv(remote_term, remote_source):
    run_cmd(remote_term, 10, f'python3 -m venv {REMOTE_SHARED}/venv')
    in_venv = f'source {REMOTE_SHARED}/venv/bin/activate && '

    run_cmd(remote_term, 30, (
        f'/bin/bash -c "{in_venv} python3 -m pip install pip wheel setuptools"'))

    run_cmd(remote_term, 60, f'/bin/bash -c "{in_venv} pip install {remote_source}"')


def _install_muscle3_native_openmpi(
        remote_source, remote_term, remote_fs, slurm_version):
    prefix = remote_fs / REMOTE_SHARED / 'muscle3-openmpi'
    prefix.mkdir()

    openmpi_hash = run_cmd(remote_term, 600, (
        '/bin/bash -c "'
        'for phash in $(/opt/spack/bin/spack find --format \\"{hash}\\" openmpi'
        '        | tr \'\\n\' \' \') ; do'
        '    if /opt/spack/bin/spack find --deps /\\${phash} |'
        f'                grep -q slurm@{slurm_version} ; then'
        '        echo \\${phash} ;'
        '     fi ;'
        'done'
        '"'))

    openmpi_version = run_cmd(remote_term, 600, (
        '/bin/bash -c "'
        f'/opt/spack/bin/spack find --format \\"{{version}}\\" /{openmpi_hash}'
        '"')).strip()

    gcc_version = _gcc_version(slurm_version)

    module_name = f'openmpi/{openmpi_version}-gcc-{gcc_version}-{openmpi_hash[:7]}'

    _logger.info(f'Slurm {slurm_version} and module {module_name}')

    run_cmd(remote_term, 600, (
        f'/bin/bash -l -c "'
        f'module load {module_name} && '
        f'cd {remote_source} && '
        f'make distclean >distclean.log 2>&1 && '
        f'PREFIX={prefix} make install >make.log 2>&1 || cat make.log"'))

    return 'openmpi', prefix, module_name


def _install_muscle3_native_intelmpi(
        remote_source, remote_term, remote_fs):
    prefix = remote_fs / REMOTE_SHARED / 'muscle3-intelmpi'
    prefix.mkdir()

    module_name = 'intel-oneapi-mpi'

    run_cmd(remote_term, 600, (
        f'/bin/bash -l -c "'
        f'module load {module_name} && '
        f'cd {remote_source} && '
        f'make distclean >distclean.log 2>&1 && '
        f'PREFIX={prefix} make install >make.log 2>&1 || cat make.log"'))

    return 'intelmpi', prefix, module_name


def _install_muscle3(local_term, repo_root, remote_term, remote_fs, slurm_version):
    remote_source = _install_remote_source(
            local_term, repo_root, remote_fs, slurm_version)
    _create_muscle3_venv(remote_term, remote_source)
    openmpi_install = _install_muscle3_native_openmpi(
            remote_source, remote_term, remote_fs, slurm_version)
    intelmpi_install = _install_muscle3_native_intelmpi(
            remote_source, remote_term, remote_fs)
    return openmpi_install, intelmpi_install


def _install_tests(repo_root, remote_term, remote_fs, remote_m3_installs):
    remote_home = remote_fs / REMOTE_SHARED

    for mpi_type, remote_m3, mpi_module in remote_m3_installs:
        cerulean.copy(
                repo_root / 'integration_test' / 'cluster_test', remote_home,
                copy_permissions=True)

        remote_source = remote_home / 'cluster_test'

        if mpi_type == 'openmpi':
            run_cmd(remote_term, 30, (
                '/bin/bash -c "'
                f'sed -i \\"s^modules: openmpi^modules: {mpi_module}^\\"'
                f' {remote_source}/implementations_openmpi.ymmsl'
                '"'))

            run_cmd(remote_term, 30, (
                '/bin/bash -c "'
                f'sed -i \\"s^modules: openmpi^modules: {mpi_module}^\\"'
                f' {remote_source}/implementations_srunmpi.ymmsl'
                '"'))

        run_cmd(remote_term, 30, (
            f'/bin/bash -l -c "'
            f'module load {mpi_module} && '
            f'. {remote_m3}/bin/muscle3.env && '
            f'make -C {remote_source} MPI_TYPE={mpi_type}"'))


def _clean_up_base_cluster(local_term, slurm_version):
    node_names = [f'node-{i}-{slurm_version}' for i in range(5)]
    run_cmd(local_term, 60, f'docker rm -f {" ".join(node_names)}')

    run_cmd(local_term, 60, f'docker rm -f headnode-{slurm_version}')

    net_name = f'muscle3-net-{slurm_version}'
    run_cmd(local_term, 60, f'docker network rm -f {net_name}')


@pytest.fixture(scope='session', params=IDX_SLURM_VERSIONS)
def installed_cluster(
        request, cleanup_docker, fake_cluster_image, fake_cluster_image_old, shared_dir,
        repo_root, local_term, assert_no_threads_left):

    slurm_version = request.param[1]
    local_shared_dir = shared_dir / slurm_version
    local_shared_dir.mkdir()
    local_shared_dir.chmod(0o1777)

    remote_term, remote_fs, headnode_port = _start_base_cluster(
            local_term, request.param, local_shared_dir)
    remote_m3_installs = _install_muscle3(
            local_term, repo_root, remote_term, remote_fs, slurm_version)
    _install_tests(repo_root, remote_term, remote_fs, remote_m3_installs)

    yield headnode_port

    # Because it's been made inside of the container, the shared directory has a
    # different owner than what we're running with on the host, and the host user cannot
    # remove the files. So we do it here from inside the container
    if CLEAN_UP_CONTAINERS:
        run_cmd(remote_term, 60, f'rm -rf {REMOTE_SHARED}/*')

    remote_fs.close()
    remote_term.close()

    if CLEAN_UP_CONTAINERS:
        _clean_up_base_cluster(local_term, slurm_version)


@pytest.fixture(scope='session')
def hwthread_to_core():
    """Translates hwthreads to core ids.

    In our tests, we use sched_getaffinity to check which cores we're bound to. This
    returns numbers identifying hwthreads, but our planner binds swthreads and processes
    to entire cores. So we get a comma-separated list of hwthread ids and want to
    compare that to a list of core ids.

    This reads /proc/cpuinfo to get the mapping between hwthreads and cores, and returns
    a function that takes a comma-separated list of hwthread ids and returns a list of
    corresponding core ids.
    """
    with open('/proc/cpuinfo', 'r') as f:
        cpuinfo = f.readlines()

    def get_values(cpuinfo, field):
        return [
                int(line.split(':')[1].strip())
                for line in cpuinfo if line.startswith(field)]

    hwthread_ids = get_values(cpuinfo, 'processor')
    core_ids = get_values(cpuinfo, 'core id')

    table = dict(zip(hwthread_ids, core_ids))

    def convert(aff_ids):
        cores = {table[i] for i in map(int, aff_ids.split(','))}
        return sorted(cores)

    return convert



import logging
from pathlib import Path
from subprocess import Popen, TimeoutExpired
import sys
from threading import Lock
from time import sleep
from typing import Dict, List, Tuple

from libmuscle.native_instantiator.agent.agent_commands import (
        CancelAllCommand, StartCommand, ShutdownCommand)
from libmuscle.native_instantiator.iagent_manager import IAgentManager
from libmuscle.native_instantiator.map_server import MAPServer
from libmuscle.native_instantiator.global_resources import global_resources
from libmuscle.planner.resources import OnNodeResources, Resources


_logger = logging.getLogger(__name__)


class AgentManager(IAgentManager):
    """Manage the node agents.

    Each node of our allocated resources gets an agent, which launches and monitors
    processes or that node. This class launches those agents across the nodes,
    and communicates with them.

    The AgentManager sits in between the NativeInstantiator and the MAPServer. It gets
    called by NativeInstantiator with requests for resources and commands to start and
    cancel processes on nodes, and it gets called by MAPServer with requests from the
    agents.
    """
    def __init__(self, agent_dir: Path) -> None:
        """Create an AgentManager.

        Create the object, then launch the agents and wait for them to connect and send
        information about the available resources.

        Args:
            agent_dir: Directory in which agents can write log files.
        """
        self._expected_nodes = global_resources().nodes
        self._nodes: Dict[str, str] = dict()
        self._resources: Resources = Resources([])
        self._resources_lock = Lock()   # protects _nodes and _resources

        self._finished_processes: List[Tuple[str, int]] = list()
        self._finished_processes_lock = Lock()

        self._server = MAPServer(self)
        self._launch_agents(agent_dir, self._server.get_location())

    def get_resources(self) -> Resources:
        """Return detected resources.

        This returns a list of sets of logical hwthread ids per core, per node.

        Called by NativeInstantiator.
        """
        # no need to lock, _resources is already in its final state
        return self._resources

    def start(
            self, node_name: str, name: str, work_dir: Path, args: List[str],
            env: Dict[str, str], stdout: Path, stderr: Path) -> None:
        """Start a process on a node.

        The files that the output is directed to will be overwritten if they already
        exist.

        Args:
            node_name: Name of the node to run the process on
            name: Name under which this process will be known
            work_dir: Working directory in which to start
            args: Executable and arguments to run
            env: Environment variables to set
            stdout: File to redirect stdout to
            stderr: File to redirect stderr to
        """
        agent_hostname = self._nodes[node_name]
        command = StartCommand(name, work_dir, args, env, stdout, stderr)
        self._server.deposit_command(agent_hostname, command)

    def cancel_all(self) -> None:
        """Cancel all processes.

        This tells the agents to stop all running processes they've started.

        Called by NativeInstantiator.
        """
        for node_host_name in self._nodes.values():
            self._server.deposit_command(node_host_name, CancelAllCommand())

    def get_finished(self) -> List[Tuple[str, int]]:
        """Returns names and exit codes of finished processes.

        This returns all processes that have finished running since the previous call;
        each started process will be returned exactly once. The names are the ones
        passed to start().

        Called by NativeInstantiator.
        """
        with self._finished_processes_lock:
            next_batch = self._finished_processes
            self._finished_processes = list()

        return next_batch

    def shutdown(self) -> None:
        """Shut down the manager and its agents."""
        command = ShutdownCommand()
        for node_host_name in self._nodes.values():
            self._server.deposit_command(node_host_name, command)

        try:
            self._agents_process.wait(60)
        except TimeoutExpired:
            _logger.warning(
                    'Agents did not shut down within one minute, sending signal...')
            self._agents_process.kill()

        try:
            self._agents_process.wait(10)
            self._agents_stdout.close()
            self._agents_stderr.close()
        except TimeoutExpired:
            _logger.warning('Agents still not down, continuing shutdown anyway.')

        self._server.stop()

    def report_resources(self, resources: OnNodeResources) -> None:
        """Report resources found on a node.

        Called by MAPServer from a server thread.

        Args:
            resources: Description of a node's resources
        """
        _logger.debug(f'Agent reported {resources}')

        # The agent reports a hostname, which we assume contains the node name
        agent_hostname = resources.node_name
        for exp_node_name in self._expected_nodes:
            if exp_node_name in resources.node_name:
                resources.node_name = exp_node_name
                break
        else:
            _logger.warning(
                    f'Agent reported hostname {resources.node_name}, which could not'
                    ' be matched against any expected node name from'
                    f' {self._expected_nodes}.')

        with self._resources_lock:
            self._nodes[resources.node_name] = agent_hostname
            self._resources.add_node(resources)

    def report_result(self, names_exit_codes: List[Tuple[str, int]]) -> None:
        """Report results of finished processes.

        Called by MAPServer from a server thread.

        Args:
            names_exit_codes: A list of names and exit codes of finished processes.
        """
        with self._finished_processes_lock:
            self._finished_processes.extend(names_exit_codes)

    def _launch_agents(self, agent_dir: Path, server_location: str) -> None:
        """Actually launch the agents.

        This runs a local process, either to start a single agent locally, or on a
        cluster to start all of them in one go.

        Args:
            agent_dir: Working directory for the agents
            server_location: MAPServer network location string for the agents to
                connect to
        """
        _logger.info('Launching MUSCLE agents...')

        python = sys.executable
        if not python:
            raise RuntimeError(
                    'Could not launch agents because sys.executable is not set.')

        log_level = logging.getLogger('libmuscle').getEffectiveLevel()

        args = [
                sys.executable, '-m', 'libmuscle.native_instantiator.agent',
                server_location, str(log_level)]

        args = global_resources().agent_launch_command(args)

        self._agents_stdout = (agent_dir / 'agent_launch.out').open('a')
        self._agents_stderr = (agent_dir / 'agent_launch.err').open('a')

        _logger.debug(f'Launching agents using {args}')
        self._agents_process = Popen(
                args, cwd=agent_dir, stdout=self._agents_stdout,
                stderr=self._agents_stderr)

        num_expected = len(self._expected_nodes)

        resources_complete = False
        while not resources_complete:
            sleep(0.1)
            with self._resources_lock:
                resources_complete = len(self._nodes) == num_expected
                too_many_agents = len(self._nodes) > num_expected

            _logger.debug(f'{len(self._nodes)} agents up of {num_expected}')

            if self._agents_process.poll() is not None:
                msg = (
                        'Agents unexpectedly stopped running. This is not supposed'
                        ' to happen. Please see the agent log for more information,'
                        ' and please file an issue on GitHub.')
                _logger.error(msg)
                raise RuntimeError(msg)

            if too_many_agents:
                msg = (
                        'More agents were started than MUSCLE3 asked for. This is not'
                        ' supposed to happen. Please file an issue on GitHub, with the'
                        ' SLURM version (use "sbatch -V") and the sbatch command used'
                        ' to submit the job.')
                _logger.error(msg)
                raise RuntimeError(msg)

        _logger.info(f'All agents running on {self._nodes}')



import pytest

from ymmsl import Conduit, Identifier as Id, Reference as Ref

from libmuscle.peer_info import PeerInfo


@pytest.fixture
def peer_info() -> PeerInfo:
    kernel = Ref('kernel')
    index = [13]

    conduits = [
            Conduit('kernel.out', 'other.in'),
            Conduit('other.out', 'kernel.in')]

    peer_dims = {
            Ref('other'): []}

    peer_locations = {
            Ref('other'): ['tcp:other']}

    return PeerInfo(kernel, index, conduits, peer_dims, peer_locations)


@pytest.fixture
def peer_info2() -> PeerInfo:
    kernel = Ref('other')
    index = []

    conduits = [
            Conduit('kernel.out', 'other.in'),
            Conduit('other.out', 'kernel.in')]

    peer_dims = {
            Ref('kernel'): [20]}

    peer_locations = {
            Ref('kernel'): ['tcp:kernel']}

    return PeerInfo(kernel, index, conduits, peer_dims, peer_locations)


@pytest.fixture
def peer_info3() -> PeerInfo:
    kernel = Ref('kernel')
    index = []

    conduits = [
            Conduit('kernel.out', 'other.in'),
            Conduit('other.out', 'kernel.in')]

    peer_dims = {
            Ref('other'): []}

    peer_locations = {
            Ref('other'): ['tcp:other']}

    return PeerInfo(kernel, index, conduits, peer_dims, peer_locations)


def test_create_peer_info(peer_info) -> None:
    assert True


def test_is_connected(peer_info) -> None:
    assert peer_info.is_connected(Id('out'))
    assert peer_info.is_connected(Id('in'))
    assert not peer_info.is_connected(Id('not_connected'))


def test_get_peer_port(peer_info, peer_info2, peer_info3) -> None:
    assert peer_info.get_peer_ports(Id('out')) == [Ref('other.in')]
    assert peer_info.get_peer_ports(Id('in')) == [Ref('other.out')]

    assert peer_info2.get_peer_ports(Id('out')) == [Ref('kernel.in')]
    assert peer_info2.get_peer_ports(Id('in')) == [Ref('kernel.out')]

    assert peer_info3.get_peer_ports(Id('out')) == [Ref('other.in')]
    assert peer_info3.get_peer_ports(Id('in')) == [Ref('other.out')]


def test_get_peer_dims(peer_info, peer_info2, peer_info3) -> None:
    assert peer_info.get_peer_dims(Ref('other')) == []
    assert peer_info2.get_peer_dims(Ref('kernel')) == [20]
    assert peer_info3.get_peer_dims(Ref('other')) == []


def test_get_peer_locations(peer_info, peer_info2, peer_info3) -> None:
    assert peer_info.get_peer_locations(Ref('other')) == ['tcp:other']
    assert peer_info2.get_peer_locations(Ref('kernel')) == ['tcp:kernel']
    assert peer_info3.get_peer_locations(Ref('other')) == ['tcp:other']


def test_get_peer_endpoint(peer_info, peer_info2, peer_info3) -> None:
    assert str(peer_info.get_peer_endpoints(Id('out'), [])[0]) == 'other.in[13]'
    assert str(peer_info.get_peer_endpoints(Id('in'), [])[0]) == 'other.out[13]'

    assert str(peer_info2.get_peer_endpoints(Id('out'), [11])[0]) == 'kernel[11].in'
    assert str(peer_info2.get_peer_endpoints(Id('in'), [11])[0]) == 'kernel[11].out'

    assert str(peer_info3.get_peer_endpoints(Id('out'), [42])[0]) == 'other.in[42]'
    assert str(peer_info3.get_peer_endpoints(Id('in'), [42])[0]) == 'other.out[42]'



import logging
from pathlib import Path
from textwrap import indent
from threading import Thread
from typing import Dict, List, Optional, Tuple, Union
from multiprocessing import Queue
from queue import Empty

from ymmsl import Configuration, Reference

from libmuscle.errors import ConfigurationError
from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.instantiator import (
        CancelAllRequest, CrashedResult, InstantiatorRequest,
        InstantiationRequest, Process, ProcessStatus, ShutdownRequest)
from libmuscle.manager.logger import last_lines
from libmuscle.manager.run_dir import RunDir
from libmuscle.native_instantiator.native_instantiator import NativeInstantiator
from libmuscle.planner.planner import Planner, ResourceAssignment
from libmuscle.planner.resources import Resources


_logger = logging.getLogger(__name__)


class LogHandlingThread(Thread):
    """Pumps log records from a queue.

    This gets log records from a queue and sends them to the logging
    system.
    """
    def __init__(self, queue: Queue) -> None:
        """Creates a LogHandlingThread.

        Args:
            queue: Queue to get log records from.
        """
        super().__init__()
        self._queue = queue
        self._shutting_down = False

    def shutdown(self) -> None:
        """Process remaining records and stop the thread."""
        self._shutting_down = True

    def run(self) -> None:
        """The thread's entry point."""
        while True:
            try:
                record = self._queue.get(True, 0.1)
                logger = logging.getLogger(record.name)
                logger.handle(record)
            except Empty:
                if self._shutting_down:
                    break


_ResultType = Union[Process, CrashedResult]


class InstanceManager:
    """Instantiates and manages running instances"""
    def __init__(
            self, configuration: Configuration, run_dir: RunDir,
            instance_registry: InstanceRegistry) -> None:
        """Create an InstanceManager.

        Args:
            configuration: The global configuration
            run_dir: Directory to run in
            instance_registry: The InstanceRegistry to use
        """
        self._configuration = configuration
        self._run_dir = run_dir
        self._instance_registry = instance_registry

        self._resources_in: Queue[Resources] = Queue()
        self._requests_out: Queue[InstantiatorRequest] = Queue()
        self._results_in: Queue[_ResultType] = Queue()
        self._log_records_in: Queue[logging.LogRecord] = Queue()

        self._instantiator = NativeInstantiator(
                self._resources_in, self._requests_out, self._results_in,
                self._log_records_in, self._run_dir.path)
        self._instantiator.start()

        self._log_handler = LogHandlingThread(self._log_records_in)
        self._log_handler.start()

        self._allocations: Optional[Dict[Reference, ResourceAssignment]] = None

        resources = self._resources_in.get()
        _logger.debug(f'Got resources {resources}')
        if isinstance(resources, CrashedResult):
            msg = (
                'Instantiator crashed. This should not happen, please file a bug'
                ' report.')
            _logger.error(msg)
            raise RuntimeError(msg) from resources.exception

        self._planner = Planner(resources)
        self._num_running = 0

    def set_manager_location(self, location: str) -> None:
        """Sets the network location of the manager.

        Call this before starting any instances so that the location
        can be passed to them.

        Args:
            location: The network location (e.g. localhost:5000)
        """
        self._manager_location = location

    def start_all(self) -> None:
        """Starts all the instances of the model."""
        self._allocations = self._planner.allocate_all(self._configuration)
        for instance, resources in self._allocations.items():
            _logger.info(f'Planned {instance} on {resources.as_resources()}')

        components = {c.name: c for c in self._configuration.model.components}
        for instance, resources in self._allocations.items():
            component = components[instance.without_trailing_ints()]
            if component.implementation is None:
                _logger.warning(
                        f'No implementation specified for {component.name}'
                        ', not starting it.')
                continue
            implementation = self._configuration.implementations[
                    component.implementation]
            implementation.env['MUSCLE_MANAGER'] = self._manager_location
            idir = self._run_dir.add_instance_dir(instance)
            workdir = idir / 'workdir'
            workdir.mkdir()
            stdout_path = idir / 'stdout.txt'
            stderr_path = idir / 'stderr.txt'

            request = InstantiationRequest(
                    instance, implementation,
                    self._configuration.resources[component.name],
                    resources, idir, workdir, stdout_path, stderr_path)
            _logger.info(f'Instantiating {instance}')
            self._requests_out.put(request)
            self._num_running += 1

    def get_resources(self) -> Dict[Reference, ResourceAssignment]:
        """Returns the resources allocated to each instance.

        Only call this after start_all() has been called, or it will raise
        an exception because the information is not available.

        Return:
            The resources for each instance instantiated by start_all()
        """
        if self._allocations is None:
            raise RuntimeError(
                    'Tried to get resources but we are running without --start-all')

        return self._allocations

    def wait(self) -> bool:
        """Waits for all instances to be done."""
        all_seemingly_okay = True

        def cancel_all() -> None:
            nonlocal all_seemingly_okay
            if all_seemingly_okay:
                self._requests_out.put(CancelAllRequest())
                all_seemingly_okay = False

        # Get all results
        results: List[Process] = list()

        while self._num_running > 0:
            result = self._results_in.get()

            if isinstance(result, CrashedResult):
                if isinstance(result.exception, ConfigurationError):
                    _logger.error(str(result.exception))
                else:
                    _logger.error(
                        'Instantiator crashed. This should not happen, please file'
                        ' a bug report.')
                return False

            results.append(result)
            if result.status != ProcessStatus.CANCELED:
                registered = self._instance_registry.did_register(result.instance)
                if result.exit_code != 0 or not registered:
                    cancel_all()
            self._num_running -= 1

        # Summarise outcome
        crashes: List[Tuple[Process, Path]] = list()
        indirect_crashes: List[Tuple[Process, Path]] = list()

        for result in results:
            if result.status == ProcessStatus.CANCELED:
                if result.exit_code == 0:
                    _logger.info(
                            f'Instance {result.instance} was not started'
                            f' because of an error elsewhere')
                else:
                    _logger.info(
                            f'Instance {result.instance} was shut down by'
                            f' MUSCLE3 because an error occurred elsewhere')
                # Ensure we don't see this as a succesful run when shutdown() is called
                # by another thread:
                all_seemingly_okay = False
            else:
                stderr_file = (
                        self._run_dir.instance_dir(result.instance) /
                        'stderr.txt')
                if result.exit_code == 0:
                    if self._instance_registry.did_register(result.instance):
                        _logger.info(
                                f'Instance {result.instance} finished with'
                                ' exit code 0')
                    else:
                        _logger.error(
                                f'Instance {result.instance} quit with no error'
                                ' (exit code 0), but it never registered with the'
                                ' manager. Maybe it never created an Instance'
                                ' object?')
                        crashes.append((result, stderr_file))
                else:
                    try:
                        with stderr_file.open() as f:
                            peer_crash = any(['peer crash?' in line for line in f])
                    except FileNotFoundError:
                        peer_crash = False

                    if peer_crash:
                        _logger.warning(
                                f'Instance {result.instance} crashed, likely because'
                                f' an error occurred elsewhere.')
                        indirect_crashes.append((result, stderr_file))
                    else:
                        _logger.error(
                                f'Instance {result.instance} quit with exit code'
                                f' {result.exit_code}')
                        crashes.append((result, stderr_file))

            _logger.debug(f'Status: {result.status}')
            _logger.debug(f'Exit code: {result.exit_code}')
            _logger.debug(f'Error msg: {result.error_msg}')

        # Show errors from crashed components
        if crashes:
            for result, stderr_file in crashes:
                _logger.error(
                        f'The last error output of {result.instance} was:')
                _logger.error(
                        '\n' + indent(last_lines(stderr_file, 20), '    '))
                _logger.error(
                        'More output may be found in'
                        f' {self._run_dir.instance_dir(result.instance)}\n'
                        )
        elif indirect_crashes:
            # Possibly a component exited without error, but prematurely. If this
            # caused ancillary crashes due to dropped connections, then the logs
            # of those will give a hint as to what the problem may be, so print
            # those instead.
            _logger.error(
                    'At this point, one or more instances crashed because they'
                    ' lost their connection to another instance, but no other'
                    ' crashing instance was found that could have caused this.')
            _logger.error(
                    'This means that either another instance quit before it was'
                    ' supposed to, but with exit code 0, or there was an actual'
                    ' network problem that caused the connection to drop.')
            _logger.error(
                    'Here is the output of the instances that lost connection:')
            for result, stderr_file in indirect_crashes:
                _logger.error(
                        f'The last error output of {result.instance} was:')
                _logger.error(
                        '\n' + indent(last_lines(stderr_file, 20), '    '))
                _logger.error(
                        'More output may be found in'
                        f' {self._run_dir.instance_dir(result.instance)}\n'
                        )
        elif not all_seemingly_okay:
            # shutdown() was called by another thread (e.g. the DeadlockDetector):
            _logger.error('The simulation was aborted.')
        else:
            _logger.info('The simulation finished without error.')

        return all_seemingly_okay

    def shutdown(self) -> None:
        """Shuts down the backend.

        This will wait for any processes still running before shutting
        down and returning.
        """
        _logger.debug('Shutting down instance manager')
        self._requests_out.put(CancelAllRequest())
        self._requests_out.put(ShutdownRequest())
        self._instantiator.join()
        self._log_handler.shutdown()
        self._log_handler.join()
        # Close multiprocessing.Queues and ensure their feeder threads exit
        queues: List[Queue] = [
                self._resources_in, self._requests_out,
                self._results_in, self._log_records_in]
        for queue in queues:
            queue.close()
            queue.join_thread()
        _logger.debug('Instance manager shut down cleanly')



import pytest
from ymmsl import Identifier

from libmuscle.port import Operator, Port


def test_create_port():
    port = Port('out', Operator.O_I, False, True, 0, [])
    assert port.name == Identifier('out')
    assert port.operator == Operator.O_I
    assert port._length is None
    assert port._is_resizable is False
    assert port._is_connected is True


def test_create_invalid_port():
    with pytest.raises(RuntimeError):
        Port('out', Operator.O_I, True, True, 1, [])

    with pytest.raises(RuntimeError):
        Port('out', Operator.O_F, True, True, 0, [10, 10])

    with pytest.raises(RuntimeError):
        Port('out', Operator.O_F, False, True, 1, [11, 12])

    with pytest.raises(RuntimeError):
        Port('out', Operator.O_I, False, True, 2, [])


def test_port_properties():
    port = Port('out', Operator.O_F, False, False, 0, [])
    assert port.is_vector() is False
    assert port.is_resizable() is False
    assert port.is_connected() is False
    with pytest.raises(RuntimeError):
        port.get_length()
    with pytest.raises(RuntimeError):
        port.set_length(3)

    port = Port('out', Operator.O_F, True, True, 0, [10])
    assert port.is_vector() is True
    assert port.is_resizable() is False
    assert port.get_length() == 10
    with pytest.raises(RuntimeError):
        port.set_length(3)

    port = Port('out', Operator.O_F, False, True, 1, [10])
    assert port.is_vector() is False
    assert port.is_resizable() is False
    with pytest.raises(RuntimeError):
        port.get_length()
    with pytest.raises(RuntimeError):
        port.set_length(4)

    port = Port('out', Operator.O_F, True, True, 1, [10, 20])
    assert port.is_vector() is True
    assert port.is_resizable() is False
    assert port.get_length() == 20
    with pytest.raises(RuntimeError):
        port.set_length(5)

    port = Port('out', Operator.O_F, False, True, 2, [12])
    assert port.is_vector() is False
    assert port.is_resizable() is False
    with pytest.raises(RuntimeError):
        port.get_length()
    with pytest.raises(RuntimeError):
        port.set_length(9)

    port = Port('out', Operator.O_F, True, True, 1, [13])
    assert port.is_vector() is True
    assert port.is_resizable() is True
    assert port.get_length() == 0
    port.set_length(27)
    assert port.get_length() == 27



from unittest.mock import patch

import msgpack
import pytest
from ymmsl import Conduit, Operator, Port, Reference

import libmuscle
from libmuscle.logging import LogLevel, LogMessage, Timestamp
from libmuscle.mcp.protocol import RequestType, ResponseType
from libmuscle.mmp_client import MMPClient


def test_init() -> None:
    with patch('libmuscle.mmp_client.TcpTransportClient') as mock_ttc:
        stub = mock_ttc.return_value
        client = MMPClient(Reference([]), '')
        assert client._transport_client == stub     # type: ignore


def test_connection_fail() -> None:
    with pytest.raises(RuntimeError):
        # Port 255 is reserved and privileged, so there's probably
        # nothing there.
        MMPClient(Reference([]), 'tcp:localhost:255')


def test_submit_log_message(mocked_mmp_client, profile_data) -> None:
    client, stub = mocked_mmp_client
    result = [ResponseType.SUCCESS.value]
    stub.call.return_value = (
            msgpack.packb(result, use_bin_type=True), profile_data)

    message = LogMessage(
            'test_mmp_client',
            Timestamp(1.0),
            LogLevel.WARNING,
            'Testing the MMPClient')

    client.submit_log_message(message)
    assert stub.call.called

    sent_request = stub.call.call_args[0][0]
    decoded_request = msgpack.unpackb(sent_request, raw=False)

    assert decoded_request == [
            RequestType.SUBMIT_LOG_MESSAGE.value,
            'test_mmp_client', 1.0, LogLevel.WARNING.value,
            'Testing the MMPClient']


def test_get_settings(mocked_mmp_client, profile_data) -> None:
    client, stub = mocked_mmp_client

    settings_msg = {
            'test1': 'test',
            'test2': 12,
            'test3': 3.14,
            'test4': True,
            'test5': [1.2, 3.4],
            'test6': [[1.2, 3.4], [5.6, 7.8]]}
    transport_result = [ResponseType.SUCCESS.value, settings_msg]
    stub.call.return_value = (
            msgpack.packb(transport_result, use_bin_type=True), profile_data)

    settings = client.get_settings()
    assert len(settings) == 6
    assert settings['test1'] == 'test'
    assert settings['test2'] == 12
    assert settings['test3'] == 3.14
    assert settings['test4'] is True
    assert settings['test5'] == [1.2, 3.4]
    assert settings['test6'] == [[1.2, 3.4], [5.6, 7.8]]


def test_register_instance(mocked_mmp_client, profile_data) -> None:
    client, stub = mocked_mmp_client

    result = [ResponseType.SUCCESS.value]
    stub.call.return_value = (
            msgpack.packb(result, use_bin_type=True), profile_data)

    client.register_instance(
            ['direct:test', 'tcp:test'],
            [Port('out', Operator.O_I), Port('in', Operator.S)])

    assert stub.call.called
    sent_msg = msgpack.unpackb(stub.call.call_args[0][0], raw=False)
    assert sent_msg == [
            RequestType.REGISTER_INSTANCE.value, 'component[13]',
            ['direct:test', 'tcp:test'], [['out', 'O_I'], ['in', 'S']],
            libmuscle.__version__]


def test_request_peers(mocked_mmp_client, profile_data) -> None:
    client, stub = mocked_mmp_client

    result_msg = [
            ResponseType.SUCCESS.value,
            [['kernel.out', 'other.in']],
            {'other': [20]},
            {'other': ['direct:test', 'tcp:test']}]
    stub.call.return_value = (
            msgpack.packb(result_msg, use_bin_type=True), profile_data)

    result = client.request_peers()

    assert stub.call.called
    sent_msg = msgpack.unpackb(stub.call.call_args[0][0], raw=False)
    assert sent_msg[0] == RequestType.GET_PEERS.value
    assert sent_msg[1] == 'component[13]'

    assert len(result[0]) == 1
    assert isinstance(result[0][0], Conduit)
    assert result[0][0].sender == 'kernel.out'
    assert result[0][0].receiver == 'other.in'

    assert isinstance(result[1], dict)
    assert result[1]['other'] == [20]

    assert isinstance(result[2], dict)
    assert result[2]['other'] == ['direct:test', 'tcp:test']


def test_request_peers_error(mocked_mmp_client, profile_data) -> None:
    client, stub = mocked_mmp_client

    result_msg = [ResponseType.ERROR.value, 'test_error_message']
    stub.call.return_value = (
            msgpack.packb(result_msg, use_bin_type=True), profile_data)

    with pytest.raises(RuntimeError):
        client.request_peers()


def test_request_peers_timeout(mocked_mmp_client, profile_data) -> None:
    client, stub = mocked_mmp_client

    result_msg = [ResponseType.PENDING.value, 'test_status_message']
    stub.call.return_value = (
            msgpack.packb(result_msg, use_bin_type=True), profile_data)

    with patch('libmuscle.mmp_client.PEER_TIMEOUT', 1), \
            patch('libmuscle.mmp_client.PEER_INTERVAL_MIN', 0.1), \
            patch('libmuscle.mmp_client.PEER_INTERVAL_MAX', 1.0):
        with pytest.raises(RuntimeError):
            client.request_peers()


def test_deregister_instance(mocked_mmp_client, profile_data) -> None:
    client, stub = mocked_mmp_client

    result = [ResponseType.SUCCESS.value]
    stub.call.return_value = (
            msgpack.packb(result, use_bin_type=True), profile_data)

    client.deregister_instance()

    assert stub.call.called
    sent_msg = msgpack.unpackb(stub.call.call_args[0][0], raw=False)
    assert sent_msg == [RequestType.DEREGISTER_INSTANCE.value, 'component[13]']


def test_deregister_instance_error(mocked_mmp_client, profile_data) -> None:
    client, stub = mocked_mmp_client

    result = [ResponseType.ERROR.value, 'Instance component[13] unknown']
    stub.call.return_value = (
            msgpack.packb(result, use_bin_type=True), profile_data)

    with pytest.raises(RuntimeError):
        client.deregister_instance()

    assert stub.call.called
    sent_msg = msgpack.unpackb(stub.call.call_args[0][0], raw=False)
    assert sent_msg == [RequestType.DEREGISTER_INSTANCE.value, 'component[13]']



import bisect
import logging
import time
from typing import List, Optional, Union

from ymmsl import (
        CheckpointRangeRule, CheckpointAtRule, CheckpointRule, Checkpoints)


_logger = logging.getLogger(__name__)


class CheckpointTrigger:
    """Represents a trigger for creating snapshots"""

    def next_checkpoint(self, cur_time: float) -> Optional[float]:
        """Calculate the next checkpoint time

        Args:
            cur_time: current time.

        Returns:
            The time when a next checkpoint should be taken, or None if this
            trigger has no checkpoint after cur_time.
        """
        raise NotImplementedError()

    def previous_checkpoint(self, cur_time: float) -> Optional[float]:
        """Calculate the previous checkpoint time

        Args:
            cur_time: current time.

        Returns:
            The time when a previous checkpoint should have been taken, or None
            if this trigger has no checkpoint after cur_time.
        """
        raise NotImplementedError()


class AtCheckpointTrigger(CheckpointTrigger):
    """Represents a trigger based on an "at" checkpoint rule

    This triggers at the specified times.
    """

    def __init__(self, at_rules: List[CheckpointAtRule]) -> None:
        """Create an "at" checkpoint trigger

        Args:
            at: list of checkpoint moments
        """
        self._at = sorted([a for r in at_rules for a in r.at])

    def next_checkpoint(self, cur_time: float) -> Optional[float]:
        if cur_time >= self._at[-1]:
            return None  # no future checkpoint left
        idx = bisect.bisect(self._at, cur_time)
        return self._at[idx]

    def previous_checkpoint(self, cur_time: float) -> Optional[float]:
        if cur_time < self._at[0]:
            return None  # no previous checkpoint
        idx = bisect.bisect(self._at, cur_time)
        return self._at[idx - 1]


class RangeCheckpointTrigger(CheckpointTrigger):
    """Represents a trigger based on a "ranges" checkpoint rule

    This triggers at a range of checkpoint moments.

    Equivalent an "at" rule ``[start, start + step, start + 2*step, ...]`` for
    as long as ``start + i*step <= stop``.

    Stop may be omitted, in which case the range is infinite.

    Start may be omitted, in which case the range is equivalent to an "at" rule
    ``[..., -n*step, ..., -step, 0, step, 2*step, ...]`` for as long as
    ``i*step <= stop``.
    """

    def __init__(self, range: CheckpointRangeRule) -> None:
        """Create a range of checkpoints

        Args:
            range: checkpoint range defining start, stop and step.
        """
        self._start = range.start
        self._stop = range.stop
        self._every = range.every
        self._last: Union[int, float, None] = None
        if self._stop is not None:
            start = 0 if self._start is None else self._start
            diff = self._stop - start
            self._last = start + (diff // self._every) * self._every

    def next_checkpoint(self, cur_time: float) -> Optional[float]:
        if self._start is not None and cur_time < self._start:
            return float(self._start)
        if self._last is not None and cur_time >= self._last:
            return None
        start = 0 if self._start is None else self._start
        diff = cur_time - start
        return float(start + (diff // self._every + 1) * self._every)

    def previous_checkpoint(self, cur_time: float) -> Optional[float]:
        if self._start is not None and cur_time < self._start:
            return None
        if self._last is not None and cur_time > self._last:
            return float(self._last)
        start = 0 if self._start is None else self._start
        diff = cur_time - start
        return float(start + (diff // self._every) * self._every)


class CombinedCheckpointTriggers(CheckpointTrigger):
    """Checkpoint trigger based on a combination of "at" and "ranges"
    """

    def __init__(self, checkpoint_rules: List[CheckpointRule]) -> None:
        """Create a new combined checkpoint trigger from the given rules

        Args:
            checkpoint_rules: checkpoint rules (from ymmsl)
        """
        self._triggers: List[CheckpointTrigger] = []
        at_rules: List[CheckpointAtRule] = []
        for rule in checkpoint_rules:
            if isinstance(rule, CheckpointAtRule):
                if rule.at:
                    at_rules.append(rule)
            elif isinstance(rule, CheckpointRangeRule):
                self._triggers.append(RangeCheckpointTrigger(rule))
            else:
                raise RuntimeError('Unknown checkpoint rule')
        if at_rules:
            self._triggers.append(AtCheckpointTrigger(at_rules))

    def next_checkpoint(self, cur_time: float) -> Optional[float]:
        checkpoints = (trigger.next_checkpoint(cur_time)
                       for trigger in self._triggers)
        # return earliest of all not-None next-checkpoints
        return min((checkpoint
                    for checkpoint in checkpoints
                    if checkpoint is not None),
                   default=None)  # return None if all triggers return None

    def previous_checkpoint(self, cur_time: float) -> Optional[float]:
        checkpoints = (trigger.previous_checkpoint(cur_time)
                       for trigger in self._triggers)
        # return latest of all not-None previous-checkpoints
        return max((checkpoint
                    for checkpoint in checkpoints
                    if checkpoint is not None),
                   default=None)  # return None if all triggers return None


class TriggerManager:
    """Manages all checkpoint triggers and checks if a snapshot must be saved.
    """

    def __init__(self) -> None:
        self._has_checkpoints = False
        self._last_triggers: List[str] = []
        self._cpts_considered_until = float('-inf')

    def set_checkpoint_info(
            self, elapsed: float, checkpoints: Checkpoints) -> None:
        """Register checkpoint info received from the muscle manager.
        """
        self._mono_to_elapsed = elapsed - time.monotonic()

        if not checkpoints:
            self._has_checkpoints = False
            return

        self._has_checkpoints = True

        self._checkpoint_at_end = checkpoints.at_end

        self._wall = CombinedCheckpointTriggers(checkpoints.wallclock_time)
        self._prevwall = 0.0
        self._nextwall: Optional[float] = self._wall.next_checkpoint(0.0)

        self._sim = CombinedCheckpointTriggers(checkpoints.simulation_time)
        self._prevsim: Optional[float] = None
        self._nextsim: Optional[float] = None

    def elapsed_walltime(self) -> float:
        """Returns elapsed wallclock_time in seconds.
        """
        return time.monotonic() + self._mono_to_elapsed

    def checkpoints_considered_until(self) -> float:
        """Return elapsed time of last should_save*
        """
        return self._cpts_considered_until

    def harmonise_wall_time(self, at_least: float) -> None:
        """Ensure our elapsed time is at least the given value
        """
        cur = self.elapsed_walltime()
        if cur < at_least:
            _logger.debug(
                    'Harmonise wall time: advancing clock by %f seconds',
                    at_least - cur)
            self._mono_to_elapsed += at_least - cur

    def should_save_snapshot(self, timestamp: float) -> bool:
        """Handles instance.should_save_snapshot
        """
        if not self._has_checkpoints:
            return False

        return self.__should_save(timestamp)

    def should_save_final_snapshot(
            self, do_reuse: bool, f_init_max_timestamp: Optional[float]
            ) -> bool:
        """Handles instance.should_save_final_snapshot
        """
        if not self._has_checkpoints:
            return False

        value = False
        if not do_reuse:
            if self._checkpoint_at_end:
                value = True
                self._last_triggers.append('at_end')
        elif f_init_max_timestamp is None:
            # No F_INIT messages received: reuse triggered on muscle_settings_in
            # message.
            _logger.debug('Reuse triggered by muscle_settings_in.'
                          ' Not creating a snapshot.')
        else:
            value = self.__should_save(f_init_max_timestamp)

        return value

    def update_checkpoints(self, timestamp: float) -> None:
        """Update last and next checkpoint times when a snapshot is made.

        Args:
            timestamp: timestamp as reported by the instance (or from incoming
                F_INIT messages for save_final_snapshot).
        """
        self._prevwall = self.elapsed_walltime()
        self._nextwall = self._wall.next_checkpoint(self._prevwall)

        self._prevsim = timestamp
        self._nextsim = self._sim.next_checkpoint(timestamp)

    def get_triggers(self) -> List[str]:
        """Get trigger description(s) for the current reason for checkpointing.
        """
        triggers = self._last_triggers
        self._last_triggers = []
        return triggers

    def __should_save(self, simulation_time: float) -> bool:
        """Check if a checkpoint should be taken

        Args:
            simulation_time: current/next timestamp as reported by the instance
        """
        if self._nextsim is None and self._prevsim is None:
            # we cannot make assumptions about the start time of a simulation,
            # a t=-1000 could make sense if t represents years since CE
            # and we should not disallow checkpointing for negative t
            previous = self._sim.previous_checkpoint(simulation_time)
            if previous is not None:
                # there is a checkpoint rule before the current moment, assume
                # we should have taken a snapshot back then
                self._nextsim = previous
            else:
                self._nextsim = self._sim.next_checkpoint(simulation_time)

        walltime = self.elapsed_walltime()
        self._cpts_considered_until = walltime

        self._last_triggers = []
        if self._nextwall is not None and walltime >= self._nextwall:
            self._last_triggers.append(f"wallclock_time >= {self._nextwall}")
        if self._nextsim is not None and simulation_time >= self._nextsim:
            self._last_triggers.append(f"simulation_time >= {self._nextsim}")
        return bool(self._last_triggers)



from unittest.mock import MagicMock, patch

import pytest

from libmuscle.mpp_server import MPPServer


@pytest.fixture(autouse=True)
def PostOffice():
    with patch('libmuscle.mpp_server.PostOffice') as PostOffice:
        yield PostOffice


@pytest.fixture
def MockTransportServer():
    MockTransportServer = MagicMock()
    transport_server = MockTransportServer.return_value
    transport_server.get_location.return_value = 'tcp:testing:9001'
    return MockTransportServer


@pytest.fixture(autouse=True)
def transport_server_types(MockTransportServer):
    with patch('libmuscle.mpp_server.transport_server_types', [MockTransportServer]):
        yield None


@pytest.fixture
def transport_server(MockTransportServer):
    return MockTransportServer.return_value


@pytest.fixture
def mpp_server():
    return MPPServer()


def test_get_locations(mpp_server, transport_server):
    assert mpp_server.get_locations() == [transport_server.get_location.return_value]



import logging
from pathlib import Path

from libmuscle.logging import LogLevel, Timestamp
from libmuscle.manager.logger import Logger


def test_log_level():
    assert LogLevel.CRITICAL.as_python_level() == logging.CRITICAL
    assert LogLevel.CRITICAL.value > LogLevel.DEBUG.value


def test_create_logger(tmpdir):
    logger = Logger(Path(str(tmpdir)))
    logger.close()


def test_log_message(logger, caplog):
    logger.log_message(
            'test_instance', Timestamp(123.0),
            LogLevel.CRITICAL, 'Testing the logging system')

    assert caplog.records[0].name == 'test_instance'
    assert caplog.records[0].levelname == 'CRITICAL'
    assert caplog.records[0].message == 'Testing the logging system'



from collections import defaultdict
from pathlib import Path
import sqlite3
import threading
from types import TracebackType
from typing import Any, cast, Dict, List, Optional, Tuple, Type, Union
from warnings import warn


class ProfileDatabase:
    """Accesses a profiling database.

    This class accesses a MUSCLE3 profiling database and provides basic
    analysis functionality.
    """
    def __init__(self, db_file: Union[str, Path]) -> None:
        """Open a ProfileDatabase.

        This opens the database file and creates a ProfileDatabase
        object that operates on it.

        Note that the connection to the database needs to be closed
        again when done. You should use this as a context manager for
        that, e.g.

        .. code-block:: python

            with ProfileDatabase('performance.sqlite') as db:
                # use db here
        """
        if not isinstance(db_file, Path):
            db_file = Path(db_file)

        self._db_file = db_file
        self._local = threading.local()

        self._smallest_timestamp: Optional[float] = None

    def close(self) -> None:
        """Close the connection to the database.

        This should be called once by each thread that used this
        object before it goes down, so that its connection to the
        database can be closed.

        It is usually better to use this class as a context manager,
        e.g.

        .. code-block:: python

            with ProfileDatabase('performance.sqlite') as db:
                # use db here
        """
        # If the thread never served a request, then we don't have a
        # connection.
        if hasattr(self._local, 'conn'):
            self._local.conn.close()

    def instance_stats(
            self) -> Tuple[List[str], List[float], List[float], List[float]]:
        """Calculate per-instance statistics.

        This calculates the total time spent computing, the total time
        spent communicating, and the total time spent waiting for
        a message to arrive, for each instance, in seconds.

        It returns a tuple of four lists, containing instance names,
        run times, communication times, and wait times. Note that the
        run times do not include data transfer or waiting for messages,
        so these three are exclusive and add up to the total time the
        instance was active.

        Note that sending messages in MUSCLE3 is partially done in the
        background. Transfer times include encoding and queueing of any
        sent messages as well as downloading received messages, but it
        does not include the sending side of the transfer, as this is
        done by a background thread in parallel with other work that is
        recorded (usually waiting, sometimes computing or sending
        another message).

        Nevertheless, this should give you an idea of which instances
        use the most resources. Keep in mind though that different
        instances may use different numbers of cores, so a model that
        doesn't spend much wallclock time may still spend many core
        hours. Secondly, waiting does not necessarily leave a core idle
        as MUSCLE3 is capable of detecting when models will not compute
        at the same time and having them share cores.

        See the profiling documentation page for an example.

        Returns:
            A list of instance names, a corresponding list of compute
            times, a corresponding list of transfer times, and a
            corresponding list of wait times.
        """
        cur = self._get_cursor()
        cur.execute("BEGIN TRANSACTION")
        cur.execute("SELECT name FROM instances")
        instances = [row[0] for row in cur.fetchall()]

        cur.execute(
                "SELECT instance, stop_time"
                " FROM all_events"
                " WHERE type = 'CONNECT'")
        start_run = dict(cur.fetchall())

        for name in instances:
            if name not in start_run:
                warn(
                        f'Instance {name} seems to have never registered with the'
                        ' manager, and will be omitted from the results. Did it crash'
                        ' on startup?')

        cur.execute(
                "SELECT instance, start_time"
                " FROM all_events"
                " WHERE type = 'SHUTDOWN_WAIT'")
        stop_run = dict(cur.fetchall())

        for name in instances:
            if name not in stop_run:
                # instances with no connected f_init ports don't have SHUTDOWN_WAIT
                # try to use the start of DISCONNECT_WAIT instead as the stop time
                cur.execute(
                        "SELECT start_time"
                        " FROM all_events"
                        " WHERE type = 'DISCONNECT_WAIT' and instance = ?", [name])
                result = cur.fetchall()
                if result:
                    stop_run[name] = result[0][0]
                else:
                    # as a last resort, just take the last registered event
                    warn(
                            f'Instance {name} did not shut down cleanly, data may be'
                            ' inaccurate or missing')

                    cur.execute(
                            "SELECT stop_time"
                            " FROM all_events"
                            " WHERE instance = ?"
                            " ORDER BY stop_time DESC LIMIT 1", [name])
                    result = cur.fetchall()
                    if result:
                        stop_run[name] = result[0][0]

        cur.execute(
                "SELECT instance, SUM(stop_time - start_time)"
                " FROM all_events"
                " WHERE type = 'SEND' OR type = 'RECEIVE'"
                " GROUP BY instance")
        comm = dict(cur.fetchall())

        cur.execute(
                "SELECT instance, SUM(stop_time - start_time)"
                " FROM all_events"
                " WHERE type = 'RECEIVE_WAIT'"
                " GROUP BY instance")
        wait = dict(cur.fetchall())

        cur.execute("COMMIT")
        cur.close()

        complete_instances = list(set(start_run) & set(stop_run))

        total_times = [(stop_run[i] - start_run[i]) * 1e-9 for i in complete_instances]
        comm_times = [
                (
                    (comm[i] if i in comm else 0) -
                    (wait[i] if i in wait else 0)
                ) * 1e-9
                for i in complete_instances]
        wait_times = [(wait[i] if i in wait else 0) * 1e-9 for i in complete_instances]
        run_times = [
                t - c - w
                for t, c, w in zip(total_times, comm_times, wait_times)]

        return complete_instances, run_times, comm_times, wait_times

    def resource_stats(self) -> Dict[str, Dict[str, float]]:
        """Calculate per-core statistics.

        This function calculates the amount of time each core has
        spent running each component assigned to it. It returns
        a dictionary indexed by `(node, core)` tuples which contains
        for each core a nested dictionary mapping components to the
        number of seconds that component used that core for. This
        includes time spent calculating and time spent receiving data,
        but not time spent waiting for input.

        Returns:
            A dictionary containing for each (node, core) tuple a
            dictionary containing for each component the total time
            it ran on that core.
        """
        instances, run_times, comm_times, _ = self.instance_stats()

        active_times = {
                i: r + c
                for i, r, c in zip(instances, run_times, comm_times)}

        cur = self._get_cursor()
        cur.execute("BEGIN TRANSACTION")
        cur.execute(
                "SELECT i.name, ac.node, ac.core"
                " FROM instances AS i"
                " JOIN assigned_cores AS ac ON (i.oid = ac.instance_oid)")
        instances_by_core = defaultdict(list)
        for name, node, core in cur.fetchall():
            instances_by_core[':'.join([node, str(core)])].append(name)

        cur.execute("COMMIT")
        cur.close()

        result: Dict[str, Dict[str, float]] = defaultdict(dict)
        for core, instances in instances_by_core.items():
            for instance in instances:
                result[core][instance] = active_times[instance]

        return result

    def time_taken(
            self, *, etype: str, instance: Optional[str] = None,
            port: Optional[str] = None, slot: Optional[int] = None,
            time: Optional[str] = 'start', etype2: Optional[str] = None,
            port2: Optional[str] = None, slot2: Optional[int] = None,
            time2: Optional[str] = 'stop', aggregate: str = 'mean') -> float:
        """Calculate time of and between events.

        This function returns the mean or total time spent on or
        between selected points in time recorded in the database, in
        nanoseconds. Note that due to operating system limitations,
        actual precision for individual measurements is limited to
        about a microsecond.

        For profiling purposes, an event is an operation performed by
        one of the instances in the simulation. It has a type, a
        start time, and a stop time. For example, when an instance
        sends a message, this is recorded as an event of type SEND,
        with associated timestamps. For some events, other information
        may also be recorded, such as the port and slot a message was
        sent or received on, message size, and so on.

        This function takes two points in time, each of which is the
        beginning or the end of a certain kind of event, and
        calculates the time between those two points. For example, to
        calculate how long it takes instance `micro` to send a message
        on port `final_state`, you can do

        .. code-block:: python

            db.time_taken(
                    etype='SEND', instance='micro', port='final_state')

        This selects events of type `SEND`, as well as an instance
        and a port, and since we didn't specify anything else, we get
        the time taken from the beginning to the end of the selected
        event. The `micro` model is likely to have sent many messages,
        and this function will automatically calculate the mean
        duration. So this tells us on average how long it takes
        `micro` to send a message on `final_state`.

        Averaging will be done over all attributes that are not
        specified, so for example if `final_state` is a vector port, then
        the average will be taken over all sends on all slots, unless
        a specific slot is specified by a `slot` argument.

        It is also possible to calculate time between different events.
        For example, if we know that `micro` receives on
        `initial_state`, does some calculations, and then sends on
        `state_out`, and we want to know how long the calculations take,
        then we can use

        .. code-block:: python

            db.time_taken(
                    instance='micro', port='initial_state',
                    etype='RECEIVE', time='stop',
                    port2='final_state', etype2='SEND',
                    time2='start')

        This gives the time between the end of a receive on
        `initial_state` and the start of a subsequent send on
        `final_state`. The arguments with a 2 at the end of their name
        refer to the end of the period we're measuring, and by default
        their value is taken from the corresponding start argument. So,
        the first command is actually equivalent to

        .. code-block:: python

            db.time_taken(
                etype='SEND', instance='micro', port='final_state',
                slot=None, time='start', etype2='SEND',
                port2='final_state', slot2=None, time2='stop')

        which says that we measure the time from the start of each send
        by `micro` on `final_state` to the end of each send on
        `final_state`, aggregating over all slots if applicable.

        Speaking of aggregation, there is a final argument `aggregate`
        which defaults to `mean`, but can be set to `sum` to calculate
        the sum instead. For example:

        .. code-block:: python

            db.time_taken(
                    etype='RECEIVE_WAIT', instance='macro',
                    port='state_in', aggregate='sum')


        gives the total time that `macro` has spent waiting for a
        message to arrive on its `state_in` port.

        If you are taking points in time from different events (e.g.
        different instances, ports, slots or types) then there must be
        the same number of events in the database for the start and
        end event. So starting at the end of `REGISTER` and stopping
        at the beginning of a `SEND` on an `O_I` port will likely not
        work, because the instance only registers once and probably
        sends more than once.

        Args:
            etype: Type of event to get the starting point from.
                Possible values: `'REGISTER'`, `'CONNECT'`, `'SHUTDOWN_WAIT'`,
                `'DISCONNECT_WAIT'`, `'SHUTDOWN'`, `'DEREGISTER'`, `'SEND'`,
                `'RECEIVE'`, `'RECEIVE_WAIT'`, `'RECEIVE_TRANSFER'`,
                `'RECEIVE_DECODE'`. See the documentation for a description
                of each.
            instance: Name of the instance to get the event from. You
                can use `%` as a wildcard matching anything. For
                example, `'macro[%'` will match all instances of the
                `macro` component, if it has many.
            port: Selected port, for send and receive events.
            slot: Selected slot, for send and receive events.
            time: Either `'start'` or `'stop'`, to select the beginning
                or the end of the specified event. Defaults to
                `'start'`.
            etype2: Type of event to get the stopping point from. See
                `etype`. Defaults to the value of `etype`.
            port2: Selected port. See `port`. Defaults to the value of
                `port`.
            slot2: Selected slot. See `slot`. Defaults to the value of
                `slot`.
            time2: Either `'start'` or `'stop'`, to select the
                beginning or the end of the specified event. Defaults
                to `'stop'`.
            aggregate: Either `'mean'` (default) or `'sum'`, to
                calculate that statistic.

        Returns:
            The mean or total time taken in nanoseconds.
        """
        if time is None:
            time = 'start'

        if etype2 is None:
            etype2 = etype

        if port2 is None:
            port2 = port

        if slot2 is None:
            slot2 = slot

        if time2 is None:
            time2 = 'stop'

        if time in ('start', 'stop'):
            timestamp = time + '_time'
        else:
            raise ValueError(
                    'Invalid time value, please specify either "start"'
                    ' or "stop".')

        if time2 in ('start', 'stop'):
            timestamp2 = time2 + '_time'
        else:
            raise ValueError(
                    'Invalid time2 value, please specify either "start"'
                    ' or "stop".')

        # To do sum(stop - start) across different events, we'd have
        # to do an expensive and tricky self-join on the events table
        # because we'd have to find the subsequent event of type 2 to
        # the current event of type 1.
        #
        # So instead, we calculate sum(stop) - sum(start), as well
        # as counts of both to make sure there's the same number and
        # also so we can calculate the average.
        #
        # We don't have enough digits in the timestamps to add up
        # more than a handful of them, so we split them and add the
        # high and low 32-bit halves separately. Note that whether
        # the right shift sign-extends is implementation-defined.
        # It does on almost all modern platforms, and anyway, you'd
        # have to time-travel to before 1970 to notice any problem.

        def get_sum_count(
                cur: sqlite3.Cursor, etype: Optional[str], timestamp: str,
                instance: Optional[str], port: Optional[str],
                slot: Optional[int]) -> Tuple[int, int, int]:
            """Get sums and count for one time point."""
            cur.execute("BEGIN TRANSACTION")
            query = (
                    "SELECT SUM({0} >> 32), SUM({0} & 0xffffffff), COUNT(*)"
                    " FROM all_events"
                    " WHERE type = ?".format(timestamp))
            qargs: List[Any] = [etype]

            if instance is not None:
                query += " AND instance = ?"
                qargs.append(instance)

            if port is not None:
                query += " AND port = ?"
                qargs.append(port)

            if slot is not None:
                query += " AND slot = ?"
                qargs.append(slot)

            cur.execute(query, qargs)
            result = cur.fetchone()
            cur.execute("COMMIT")
            return cast(Tuple[int, int, int], result)

        cur = self._get_cursor()
        sum1_high, sum1_low, count1 = get_sum_count(
                cur, etype, timestamp, instance, port, slot)
        sum2_high, sum2_low, count2 = get_sum_count(
                cur, etype2, timestamp2, instance, port2, slot2)
        cur.close()

        if count1 != count2:
            raise RuntimeError(
                    'The number of start and stop events is not the same:'
                    f' {count1} != {count2}')

        if count1 == 0:
            if aggregate == 'sum':
                return 0.0
            raise RuntimeError('No matching events were found')

        diff_high = (sum2_high - sum1_high) << 32
        diff_low = sum2_low - sum1_low
        if aggregate == 'mean':
            return (diff_high + diff_low) / count1
        elif aggregate == 'sum':
            return diff_high + diff_low
        else:
            raise ValueError(
                    f'Unknown aggregate {aggregate}, please specify either'
                    ' "mean" or "sum"')

    def __enter__(self) -> 'ProfileDatabase':
        return self

    def __exit__(
            self, exc_type: Optional[Type[BaseException]],
            exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
            ) -> None:
        self.close()

    def _get_cursor(self) -> sqlite3.Cursor:
        """Get a connection to run queries with.

        Exactly how the sqlite3 Python module's automatic
        transactions work is a bit mysterious, and the documentation
        of sqlite itself isn't perfect either. So we set
        isolation_level to None, which doesn't actually affect the
        effective isolation level (!) but does keep the Python module
        from starting transactions automatically, and then we use
        explicit BEGIN TRANSACTION statements to keep the sqlite
        library from doing anything automatically. That way, it's
        clear what's going on from the code.

        Also, connections are expensive and cannot be used in other
        threads. We have a multithreaded TCP server in the manager,
        so we use thread-local storage to get a connection in each
        thread.
        """
        if not hasattr(self._local, 'conn'):
            self._local.conn = sqlite3.connect(
                    self._db_file, isolation_level=None)
        return cast(sqlite3.Cursor, self._local.conn.cursor())
